/**
 * Copyright (c) 2013 Tobias Mller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using UnityEngine;
using System.Collections;
using System;

/**
 * The ARConsoleButton is the logical representation of a button on the AR console.
 * It also holds the visual representation and is depentend on the marker object.
 * It has several visual functionality such as:
 * - Pressed/Unpressed: Changing appearance when pressed.
 * - Active/Not Active: Changing appearance when not active.
 * - Highlighted: Whether the button should be highlighted to show the user which button he should press.
 * - ConsoleCommandType: The type of command being invoked when pressing the button.
 */
[RequireComponent(typeof(MarkerObject))]
public class ARConsoleButton : MonoBehaviour 
{
    MarkerObject markerObject;
    ARConsoleManager arConsoleManager;

    [SerializeField] private GameObject buttonPrefab;
    [SerializeField] private ARConsoleManager.ConsoleCommandType commandType;

    public ARConsoleManager.ConsoleCommandType CommandType { get { return commandType; } }

    private GameObject buttonObject;
    private GameObject hightlightButtonObject;
    
    /**
     * Accumulated time, the buttons has not moved
     */
    private float timeNotMoved = 0;
    private bool pushable = true;
    private bool pushed = false;
    private bool buttonActive = true;

    public bool ButtonActive 
    { 
        get { return buttonActive; } 
        set
        {
            buttonActive = value;
            UpdateStateDependentAppearance();
        }
    }

    private bool highlighted = false;
    public bool Highlighted 
    {
        get { return highlighted; }
        set
        {
            highlighted = value;
            UpdateStateDependentAppearance();
        }
    }

	void Awake () 
    {
        markerObject = GetComponent<MarkerObject>();
        arConsoleManager = ARConsoleManager.FindInstance();

        // Button not pressed prefab
        if (buttonPrefab)
        {
            buttonObject = (GameObject)Instantiate(buttonPrefab, transform.position, buttonPrefab.transform.rotation);
            buttonObject.transform.parent = transform;
            buttonObject.transform.localPosition = buttonPrefab.transform.localPosition;
        }
        else if (arConsoleManager.DefaultButtonPrefab)
        {
            buttonObject = (GameObject)Instantiate(arConsoleManager.DefaultButtonPrefab, transform.position, arConsoleManager.DefaultButtonPrefab.transform.rotation);
            buttonObject.transform.parent = transform;
        }
                
        // Button highlighted prefab
        if (arConsoleManager.DefaultHighlightButtonPrefab)
        {
            hightlightButtonObject = (GameObject)Instantiate(arConsoleManager.DefaultHighlightButtonPrefab, transform.position, arConsoleManager.DefaultHighlightButtonPrefab.transform.rotation);
            hightlightButtonObject.transform.localPosition = new Vector3(transform.localPosition.x, transform.localPosition.y + 3, transform.localPosition.z);
            hightlightButtonObject.transform.localScale = new Vector3(0.3f,0.3f,0.3f);
            hightlightButtonObject.transform.parent = transform;
        }
	}

    void Start()
    {
        if (markerObject == null)
            return;

        markerObject.MakeRequired(false);

        markerObject.EventPositionChanged += ButtonPositionChanged;

        markerObject.EventDetectionStarted += DetectionStarted;
        markerObject.EventDetectionEnded += DetectionEndet;

        UpdateStateDependentAppearance();
    }
        
    /**
     * Gets called when the button has moved. Makes the Button inactive.
     */
    private void ButtonPositionChanged(MarkerObject markerObject)
    {
        pushable = false;
        timeNotMoved = 0;
        UpdateStateDependentAppearance();
    }

    private void DetectionEndet(MarkerObject markerObject)
    {
        pushed = true;
        UpdateStateDependentAppearance();
    }

    private void DetectionStarted(MarkerObject markerObject)
    {
        pushed = false;
        UpdateStateDependentAppearance();
    }

    /**
     * This method gets called if any values are changed that affect the buttons appearance.
     */
    public void UpdateStateDependentAppearance()
    {
        buttonObject.SetActive(arConsoleManager.ConsoleActive && ButtonActive);
        hightlightButtonObject.SetActive(arConsoleManager.ConsoleActive && Highlighted);

        if (pushable && pushed)
        {
            ChangeAlphaOfButtonObject(0.5f);
        }
        else
        {
            ChangeAlphaOfButtonObject(1.0f);
        }
    }

    /**
     * Changes the alpha value of this button making it transparent.
     */
    private void ChangeAlphaOfButtonObject(float alpha)
    {
        // Changing Alpha of button mesh.
        if (buttonObject.GetComponent<MeshRenderer>() != null)
        {
            ChangeAlphaOfMeshRenderer(buttonObject.GetComponent<MeshRenderer>(), alpha);
        }
        // If the button mesh has no MeshRenderer attached, the childrens MeshRenderer is used.
        else
        {
            foreach (var meshRenderer in buttonObject.GetComponentsInChildren<MeshRenderer>())
	        {
                ChangeAlphaOfMeshRenderer(meshRenderer, alpha);
	        }
        }
    }

    /**
     * Changes the alpha value of the given MeshRenderer. This makes the mesh transparent.
     */
    private void ChangeAlphaOfMeshRenderer(MeshRenderer meshRenderer, float alpha)
    {
        var color = meshRenderer.material.color;
        color.a = alpha;
        meshRenderer.material.color = color;
    }
	
	void Update () 
    {
        // If the button is pushable, count the time it is not moved.
        // When it reaches the ReactionDelayAfterMoving border, the button is pushable again.
        if (!pushable)
        {
            timeNotMoved += Time.deltaTime;
            if (timeNotMoved > arConsoleManager.ReactionDelayAfterMoving)
            {
                pushable = true;
                UpdateStateDependentAppearance();
            }
        }
        // Rotate the highlighting indicator mesh, if the button is currently highlighted.
        if (Highlighted)
        {
            hightlightButtonObject.transform.Rotate(Vector3.up, Time.deltaTime * 250);
        }
	}
}
