/**
 * Copyright (c) 2013 Tobias Mller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using UnityEngine;
using System.Collections;
using System;
using System.Collections.Generic;
using NyARUnityUtils;
using System.IO;
using System.Linq;

/**
 * The ARConsoleManager is the core service for interaction and usage of the Console consisting of AR markers.
 * Other classes can use this singleton class to substricbe to certain events and commands. The main events are the following:
 * - ConsoleActivatedEvent: This event is triggered when the console becomes completely visible to the camera.
 * - ConsoleDeactivatedEvent: This event is triggered when the console is completely invisible to the camera.
 * 
 * The ARConsoleManager holds several measuring settings to configure the console to the cameras likings.
 * 
 * The specific console command events are the following. They can be used to...
 * - IncrementSize: ...increment the playFields size.
 * - DecrementSize: ...decrement the playFields size.
 * - ToggleCalibration: ...toggle the calibration process prio to the game.
 * - FlipCameraHorizontal: ...flip the camera horizontal to the users likings.
 * - FlipCameraVertical: ...flip the camera vertical to the users likings.
 * - RestartGame: ...start a completely new game.
 * - ToggleSpeaker: ...Mute/unmute/lessen the musics volume
 */
public class ARConsoleManager : MonoBehaviour
{
    private ARManager arManager;
    public delegate void OnCommand();

    /**
     * The command types other Components of the game can subscribe to.
     */
    public enum ConsoleCommandType
    {
        IncrementSize,
        DecrementSize,
        ToggleCalibration,
        FlipCameraHorizontal,
        FlipCameraVertical,
        RestartGame,
        ToggleSpeaker
    }

    private List<ARConsoleButton> arButtons;

    // Default values for buttons appearance
    [SerializeField] private GameObject defaultButtonPrefab;
    public GameObject DefaultButtonPrefab { get { return defaultButtonPrefab; } }  
    [SerializeField] private GameObject defaultHighlightButtonPrefab;
    public GameObject DefaultHighlightButtonPrefab { get { return defaultHighlightButtonPrefab; } }

    [SerializeField] private float reactionDelayAfterMoving = 1;
    /**
     * The delay a button needs to be stationary to be pushable again
     */
    public float ReactionDelayAfterMoving { get { return reactionDelayAfterMoving; } }

    private Dictionary<ConsoleCommandType, OnCommand> consoleCommands;
    /**
     * The command types Dictionary other Components of the game can easily subscribe to.
     */
    private Dictionary<ConsoleCommandType, OnCommand> ConsoleCommands { get { return consoleCommands; } }

    /**
     * This Event gets fired when the console becomes completely visible
     */
    public OnCommand ConsoleActivatedEvent;

    /**
     * This Event gets fired when the console becomes completely invisible
     */
    public OnCommand ConsoleDeactivatedEvent;

    private bool consoleActive;

    /**
     * Accumulated time, the whole console has not moved
     */
    private float timeNotMoved;

    private bool consoleMoving;

    /**
     * This value stores whether the console is currently active. 
     * Changing the value raises corresponding events.
     */
    public bool ConsoleActive
    {
        get { return consoleActive; }
        set
        {
            if (value == consoleActive)
            {
                return;
            }

            if (value)
            {
                if (ConsoleActivatedEvent != null)
                {
                    ConsoleActivatedEvent();
                }
            }
            else
            {
                if (ConsoleDeactivatedEvent != null)
                {
                    ConsoleDeactivatedEvent();
                }
            }

            consoleActive = value;
            arButtons.ForEach(button => button.UpdateStateDependentAppearance());
        }
    }

    /**
     * Returns whether any buttons of the console can be pushed
     */
    public bool ButtonsPushable { get { return ConsoleActive && !consoleMoving; } }

    void Awake()
    {
        arButtons = new List<ARConsoleButton>();
        consoleCommands = new Dictionary<ConsoleCommandType, OnCommand>();

        arManager = ARManager.FindInstance();
        arManager.CallWhenInitialized(Initialize);
    }

    private void Initialize()
    {
        // Go through all buttons and add them to arButtons. Subscribe to button events to update console status.
        foreach (var markerObject in arManager.MarkerObjects)
        {
            var arConsoleButton = markerObject.GetComponent<ARConsoleButton>();
            if (arConsoleButton != null)
            {
                arButtons.Add(arConsoleButton);
                markerObject.EventDetectionStarted += ButtonDetectionStarted;
                markerObject.EventDetectionEnded += ButtonDetectionEnded;
                markerObject.EventPositionChanged += ButtonPositionChanged;

                if (!ConsoleCommands.ContainsKey(arConsoleButton.CommandType))
                {
                    ConsoleCommands.Add(arConsoleButton.CommandType, null); // null = event not used yet
                }
            }
        }

        ConsoleActive = (arButtons.Count(button => button.GetComponent<MarkerObject>().Visible) == arButtons.Count);
	}

    public void AddConsoleCommand(ConsoleCommandType commandType, OnCommand callback)
    {
        if (!ConsoleCommands.ContainsKey(commandType))
        {
            ConsoleCommands[commandType] = callback;
        }
        else
        {
            ConsoleCommands[commandType] += callback;
        }
    }

    public void DeactivateButton(ConsoleCommandType type)
    {
        GetButton(type).ButtonActive = false;
    }

    public void ActivateButton(ConsoleCommandType type)
    {
        GetButton(type).ButtonActive = true;
    }

    public void HighlightButton(ConsoleCommandType type, bool highlight)
    {
        GetButton(type).Highlighted = highlight;
    }

    public ARConsoleButton GetButton(ConsoleCommandType type)
    {
        foreach (var button in arButtons)
        {
            if (button.CommandType == type)
            {
                return button;
            }
        }
        return null;
    }

    private void ButtonPositionChanged(MarkerObject markerObject)
    {
        consoleMoving = true;
        timeNotMoved = 0;
    }

    private void ButtonDetectionEnded(MarkerObject markerObject)
    {
        var buttonsPressed = arButtons.Count(button => !button.GetComponent<MarkerObject>().Visible);
        // If the new pressed button is the only one pressed, invoke its command.
        if (ButtonsPushable && buttonsPressed == 1)
        {
            var arConsoleButton = markerObject.GetComponent<ARConsoleButton>();
            if (arConsoleButton.ButtonActive && ConsoleCommands[arConsoleButton.CommandType] != null)
                ConsoleCommands[arConsoleButton.CommandType]();
        }

        // If all buttons are not visible, deactivate the console.
        if (buttonsPressed == arButtons.Count())
        {
            ConsoleActive = false;
        }
    }

    private void ButtonDetectionStarted(MarkerObject markerObject)
    {
        // If all buttons are visible again, make the console active.
        if(arButtons.Count(button => !button.GetComponent<MarkerObject>().Visible) == 0)
        {
            ConsoleActive = true;
        }
    }
    	
	void Update()
    {
        if (!arManager.Initialized)
            return;

        if (consoleMoving)
        {
            timeNotMoved += Time.deltaTime;
            if (timeNotMoved > reactionDelayAfterMoving)
            {
                consoleMoving = false;
            }
        }
	}

    /**
     * Singleton: Returns the instance of the ARConsoleManager
     */
    public static ARConsoleManager FindInstance()
    {
        return UnityHelper.FindFirstActiveInstance<ARConsoleManager>();
    }
}
