/**
 * Copyright (c) 2013 Tobias Mller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using System;
using UnityEngine;
using System.Collections;

/*
 * A MarkerObject is the virtual positioned counterpart to the physical real-world marker. The ARManager
 * updates it regularely. It provides the following functionality:
 * 
 * - Information: It gives information about whether it is detected and where.
 * - Requiredness/Moveability: The MarkerObject can be set to required and additionally be said to be,
 *   for example, only be moved in a certain range. If a marker is required, it will never truely vanish
 *   but keep its last seen position, but it will be marked Illegal.
 * - Snapping: MarkerObjects can be snapped to the plane of the Origin Marker (see ARManager)
 * - Events: MarkerObjects throw events when newly detected, updated or the detection ended.
 */
public class MarkerObject : MonoBehaviour
{
    [SerializeField] private bool snapToOriginPlane;
    [SerializeField] private bool debugShowData;

    // This is the position/rotation the marker is actually detected on (possibly with snapping though)
    // Normally that's the same as transform.position and transform.rotation, but if the object is immovable
    // transform.position/rotation doesn't change.
    public Vector3 DetectedLocalPosition { get; protected set; }
    public Quaternion DetectedLocalRotation { get; protected set; }

    public bool Detected { get; private set; }

    // If the MarkerObject is Required, it will stay on screen on its last found position even if its not
    // detected anymore.
    public bool Required { get; protected set; }

    // Is the object moveable? Freely? In a certain radius? "Not freely moveable" only makes sense if the
    // marker is Required, otherwise somebody could simply hide it.
    public bool Moveable { get; private set; }
    public Vector3? MoveableCenter { get; private set; }
    public float MoveableRadius { get; private set; }

    public MarkerState State { get; private set; }

    // Should the marker try to snap to the origin plane?
    public bool SnapToOriginPlane { get { return snapToOriginPlane; } }

    // Is this marker currently snapped to the origin plane?
    public bool SnappedToOriginPlane { get; private set; }

    public ARManager.MarkerModelPair MarkerModelPair { get; set; }

    //public Vector3 offset = new Vector3();

    public event ARManager.DetectionChange EventDetectionStarted;
    public event ARManager.DetectionChange EventDetectionUpdated;
    public event ARManager.DetectionChange EventDetectionEnded;

    public Action<MarkerObject> EventPositionChanged;

    private ARManager arManager;
    private FlipCamera flipCamera;

    // Visual helpers to denote marker status. This was needed in an earlier prototype
    // for a different game and left in to make the framework more generally useable.
    private GameObject legalIndicator;
    private GameObject illegalIndicator;
    private GameObject requiredNotMovableIndicator;
    private GameObject correctIllegalPositionHelper;

    private GameObject markerMovementCircleLegalPrefab;
    private GameObject markerMovementCircleIllegalPrefab;
    private GameObject markerMovementPerimeterPrefab;
    private float markerMovementCircleRadius;

    private GameObject markerMovementCircleLegal;
    private GameObject markerMovementCircleIllegal;
    private GameObject markerMovementPerimeter;

    void Awake()
	{
	    Detected = false;
	    Required = false;
        Moveable = true;

	    //MoveableCenter = null;
	    //MoveableRadius = 0;
        State = MarkerState.NotFound;

        Hide();

        arManager = ARManager.FindInstance();
        flipCamera = FlipCamera.FindInstance();
	}

    void Start()
    {
        // Create helper objects
        var data = arManager.MarkerObjectConfigData;
        if (data.legalMarkerIndictatorPrefab)
        {
            legalIndicator = (GameObject) Instantiate(data.legalMarkerIndictatorPrefab, transform.position, transform.rotation);
            legalIndicator.transform.parent = transform;
            legalIndicator.SetActive(false);
        }

        if (data.illegalMarkerIndictatorPrefab)
        {
            illegalIndicator = (GameObject) Instantiate(data.illegalMarkerIndictatorPrefab, transform.position, transform.rotation);
            illegalIndicator.transform.parent = transform;
            illegalIndicator.SetActive(false);
        }

        if (data.requiredNotMovableIndicatorPrefab)
        {
            requiredNotMovableIndicator = (GameObject) Instantiate(data.requiredNotMovableIndicatorPrefab, transform.position, transform.rotation);
            requiredNotMovableIndicator.transform.parent = transform;
            requiredNotMovableIndicator.SetActive(false);
        }

        if (data.correctIllegalPositionHelperPrefab)
        {
            correctIllegalPositionHelper = (GameObject) Instantiate(data.correctIllegalPositionHelperPrefab, transform.position, transform.rotation);
            correctIllegalPositionHelper.transform.parent = transform;
            correctIllegalPositionHelper.SetActive(false);
        }

        markerMovementCircleLegalPrefab = data.markerMovementCircleLegalPrefab;
        markerMovementCircleIllegalPrefab = data.markerMovementCircleIllegalPrefab;
        markerMovementPerimeterPrefab = data.markerMovementPerimeterPrefab;
        markerMovementCircleRadius = data.markerMovementCircleRadius;
    }

    // This is called when the markers are updated and we actually found this marker
    public void DetectedUpdate(Transform newTransform, bool snappedToOriginPlane)
    {
        SnappedToOriginPlane = snappedToOriginPlane;

        DetectedLocalPosition = newTransform.localPosition;
        DetectedLocalRotation = newTransform.localRotation;

        if (CanBeMoved)
        {
            if (Vector3.Distance(transform.position, newTransform.position) > arManager.MovementDetectionBorder)
            {
                if (EventPositionChanged != null)
                {
                    EventPositionChanged(this);
                }
            }

            // Move actual object if movement is allowed at all
            // We do not care if the snap is wrong - that will be taken care of by MarkerState.IllegalPosition
            transform.position = newTransform.position;
            transform.rotation = newTransform.rotation;
        }

        Detected = true;

        UpdateState();

        if (correctIllegalPositionHelper && correctIllegalPositionHelper.activeSelf)
        {
            correctIllegalPositionHelper.transform.position = DetectedLocalPosition;
            correctIllegalPositionHelper.transform.rotation = DetectedLocalRotation;
        }
    }

    // This is called when the markers are updated and we didn't find this marker
    public void NotDetectedUpdate()
    {
        if (!Detected)
            return;

        Detected = false;

        if (!Required)
        {
            Hide();
        }

        UpdateState();
    }

    private void Hide()
    {
        transform.position = new Vector3(0, 0, -100);
    }

    public void MakeRequired(bool makeImmovable)
    {
        Required = true;

        if (makeImmovable)
            SetImmovable();

        UpdateState();
    }

    public void MakeUnrequired()
    {
        Required = false;
        UpdateState();
    }

    public void SetImmovable()
    {
        if (!Required)
            throw new Exception("It makes no sense to change the movability when the MarkerObject is not required");

        Moveable = false;
        DeactivateMoveableInRadiusIfActive();

        UpdateState();
    }

    public void SetFreelyMoveable()
    {
        if (!Required)
            throw new Exception("It makes no sense to change the movability when the MarkerObject is not required");

        Moveable = true;
        DeactivateMoveableInRadiusIfActive();

        UpdateState();
    }

    public void SetMoveableCircle(float radius)
    {
        if (!Required)
            throw new Exception("It makes no sense to change the movability when the MarkerObject is not required");

        Moveable = true;

        MoveableCenter = transform.localPosition;
        MoveableRadius = radius;

        if (markerMovementCircleLegal == null)
        {
            if (markerMovementCircleLegalPrefab)
            {
                markerMovementCircleLegal = (GameObject) Instantiate(markerMovementCircleLegalPrefab, Vector3.zero, transform.rotation);
                markerMovementCircleLegal.transform.parent = transform;
                markerMovementCircleLegal.transform.localPosition = Vector3.zero;
                markerMovementCircleLegal.transform.localScale = new Vector3(markerMovementCircleRadius,
                                                                             markerMovementCircleLegal.transform.localScale.y,
                                                                             markerMovementCircleRadius);
            }

            if (markerMovementCircleIllegalPrefab)
            {
                markerMovementCircleIllegal = (GameObject) Instantiate(markerMovementCircleIllegalPrefab, Vector3.zero, transform.rotation);
                markerMovementCircleIllegal.transform.parent = transform;
                markerMovementCircleIllegal.transform.localPosition = Vector3.zero;
                markerMovementCircleIllegal.transform.localScale = new Vector3(markerMovementCircleRadius,
                                                                               markerMovementCircleIllegal.transform.localScale.y,
                                                                               markerMovementCircleRadius);
            }

            if (markerMovementPerimeterPrefab)
            {
                markerMovementPerimeter = (GameObject) Instantiate(markerMovementPerimeterPrefab, Vector3.zero, transform.rotation);
                markerMovementPerimeter.transform.parent = transform.parent;
            }
        }

        float shownRadius = markerMovementCircleRadius + radius;

        if (markerMovementCircleLegal != null)
            markerMovementCircleLegal.SetActive(true);

        if (markerMovementCircleIllegal != null)
            markerMovementCircleIllegal.SetActive(false);

        if (markerMovementPerimeter != null)
        {
            markerMovementPerimeter.SetActive(true);
            markerMovementPerimeter.transform.localPosition = MoveableCenter.Value;
            markerMovementPerimeter.transform.localScale = new Vector3(shownRadius, markerMovementPerimeter.transform.localScale.y, shownRadius);
        }

        UpdateState();
    }

    private void DeactivateMoveableInRadiusIfActive()
    {
        if (MoveableCenter == null)
            return;

        MoveableCenter = null;
        //markerMovementCircle.SetActive(false);
        //markerMovementPerimeter.SetActive(false);

        Destroy(markerMovementCircleLegal);
        markerMovementCircleLegal = null;

        Destroy(markerMovementCircleIllegal);
        markerMovementCircleIllegal = null;

        Destroy(markerMovementPerimeter);
        markerMovementPerimeter = null;
    }

    private void UpdateState()
    {
        if (Detected)
        {
            var positionLegal = CheckDetectedPositionLegality();
            var snapLegal = CheckSnapLegality();

            State = (positionLegal && snapLegal)
                        ? MarkerState.Recognized
                        : MarkerState.IllegalPosition;

            if (correctIllegalPositionHelper)
                correctIllegalPositionHelper.SetActive(!positionLegal && !MoveableInRadius);
        }
        else
        {
            State = Required
                        ? MarkerState.RequiredAndNotFound
                        : MarkerState.NotFound;

            if (correctIllegalPositionHelper)
                correctIllegalPositionHelper.SetActive(false);
        }

        if (legalIndicator)
            legalIndicator.SetActive(Legal && !MoveableInRadius);

        if (illegalIndicator)
            illegalIndicator.SetActive(!Legal && !MoveableInRadius);

        if (requiredNotMovableIndicator)
            requiredNotMovableIndicator.SetActive(Required && !CanBeMoved);

        if (markerMovementCircleLegal)
            markerMovementCircleLegal.SetActive(Legal && MoveableInRadius);

        if (markerMovementCircleIllegal)
            markerMovementCircleIllegal.SetActive(!Legal && MoveableInRadius);
    }

    private bool CheckDetectedPositionLegality()
    {
        if (CanBeMoved)
        {
            if (MoveableFreely)
                return true;

            if (MoveableInRadius)
                return GetDetectedPositionDistanceFromMovementCenter() <= MoveableRadius;

            throw new Exception("A marker object claims to be able to move, but none of the options fit. There seems to be missing code here.");
        }

        if (GetDistanceFromDetectedPosition() > arManager.MaxImmovableMarkerPositionOffset)
            return false;

        if (GetDistanceFromDetectedRotation() > arManager.MaxImmovableMarkerRotationOffset)
            return false;

        return true;
    }

    private bool CheckSnapLegality()
    {
        return (!SnapToOriginPlane || SnappedToOriginPlane);
    }

    private float GetDistanceFromDetectedPosition()
    {
        return Vector3.Distance(transform.localPosition, DetectedLocalPosition);
    }

    private float GetDistanceFromDetectedRotation()
    {
        return Quaternion.Angle(transform.localRotation, DetectedLocalRotation);
    }

    private float GetDetectedPositionDistanceFromMovementCenter()
    {
        if (MoveableCenter == null)
            return float.PositiveInfinity;

        return Vector3.Distance(MoveableCenter.Value, DetectedLocalPosition);
    }

    public bool CanBeMoved { get { return !Required || Moveable; } }
    public bool AlreadyAdded { get { return Detected || Required; } }
    public bool Legal { get { return ((State == MarkerState.Recognized) || (State == MarkerState.NotFound)); } }
    public bool Visible { get { return (State == MarkerState.IllegalPosition || State == MarkerState.Recognized); } }

    public bool MoveableFreely { get { return CanBeMoved && (MoveableCenter == null); } }
    public bool MoveableInRadius { get { return CanBeMoved && (MoveableCenter != null); } }

    public void DispatchEventDetectionStarted()
    {
        if (EventDetectionStarted != null)
            EventDetectionStarted(this);
    }

    public void DispatchEventDetectionUpdated()
    {
        if (EventDetectionUpdated != null)
            EventDetectionUpdated(this);
    }

    public void DispatchEventDetectionEnded()
    {
        if (EventDetectionEnded != null)
            EventDetectionEnded(this);
    }

    public bool DebugShowData
    {
        get { return debugShowData; }
        set { debugShowData = value; }
    }

    // Output lots of debug data on screen next to the marker if requested
    void OnGUI()
    {
        if (!DebugShowData)
            return;

        string text = State.ToString();

        text += "\n" + transform.localPosition;
        text += "\n" + transform.localRotation;

        if (Required)
        {
            text += "\nRequired, ";

            if (Moveable)
            {
                text += "Moveable";

                if (MoveableInRadius)
                {
                    text += "\nCurrent Distance: " + GetDetectedPositionDistanceFromMovementCenter();
                }
            }
            else
            {
                text += "Immovable:\n" +
                    "Distance: " + GetDistanceFromDetectedPosition() + "\n" +
                    "Angle: " + GetDistanceFromDetectedRotation();
            }
        }

        if (SnapToOriginPlane)
        {
            if (SnappedToOriginPlane)
            {
                text += "\n[x] Snapped";
            }
            else
            {
                text += "\n[ ] Snapped";
            }
        }

        text += "\nMarker: " + MarkerModelPair.MarkerPattern.name;

        if (text.Length == 0)
            return;

        Vector3 worldPosition = transform.position;// +new Vector3(0, 1f, 0);

        Vector3 viewportPosition = Camera.main.WorldToViewportPoint(worldPosition);
        if ((viewportPosition.x < 0) || (viewportPosition.x > 1) ||
            (viewportPosition.y < 0) || (viewportPosition.y > 1) ||
            (viewportPosition.z < 0))
            return;

        GUI.skin = null;

        Vector2 position = Camera.main.WorldToScreenPoint(worldPosition);

        position.y = Screen.height - position.y;

        position = flipCamera.TransformScreenPosition(position);

        position.y -= 40;

        Vector2 size = GUI.skin.box.CalcSize(new GUIContent(text));
        Rect rect = new Rect(position.x - size.x / 2, position.y - size.y, size.x, size.y);

        GUI.skin.box.alignment = TextAnchor.MiddleCenter;
        GUI.Box(rect, text);
    }
}
