﻿/**
 * Copyright (c) 2013 Tobias Müller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using UnityEngine;
using System.Text;

/*
 * A LaserThread is a single connected string of positions symbolizing a laser. It has a power and a color, and
 * its tree-like structure allows it to be split in one or multiple children (nextThreads).
 * 
 * It also stores which entities came in contact with the thread (if the Entity wants that to happen), and it
 * can store entities or Transforms to be ignore in the next calculation step.
 */
public class LaserThread
{
    public float Power { get; private set; }
    public FlexibleColor Color { get; private set; }

    // A string of positions denoting the path of the thread
    private List<Vector3> positions;

    // The next threads splitting up from this. For example the Prisma splits a white thread in three colored threads
    private List<LaserThread> nextThreads;

    // Sometimes an ingame object consists of multiple parts, but the thread should pass through thos once it touched some
    // part of the ingame object. Those (or the whole entity) can be ignored for the next LaserEmitter emitting step, effectively
    // pretending the object just isn't there anymore.
    private List<Transform> ignoreTransformsUntilNextStep;
    private Entity ignoreEntityUntilNextStep;

    // A list entities might add themselves too to find out if this particular LaserThread (or its parent) hit them already
    private List<LaserReceiverMonoBehaviour> usedEntities;

    // CurrentPosition is always the last entry in positions, and CurrentDirection is used while calculating the laser. The laser
    // is calculated until CurrentDirection is Vector3.Zero
    public Vector3 CurrentDirection { get; set; }
    public Vector3 CurrentPosition { get; private set; }

    public ReadOnlyCollection<Vector3> Positions { get { return positions.AsReadOnly(); } }
    public ReadOnlyCollection<LaserThread> NextThreads { get { return nextThreads.AsReadOnly(); } }
    public ReadOnlyCollection<LaserReceiverMonoBehaviour> UsedEntities { get { return usedEntities.AsReadOnly(); } }

    public LaserThread(Vector3 startPosition, Vector3 startDirection, float power, FlexibleColor color, IEnumerable<LaserReceiverMonoBehaviour> usedEntities)
    {
        positions = new List<Vector3>();
        nextThreads = new List<LaserThread>();

        this.usedEntities = (usedEntities == null)
                                ? new List<LaserReceiverMonoBehaviour>()
                                : new List<LaserReceiverMonoBehaviour>(usedEntities);

        AddPosition(startPosition, startDirection);

        Power = power;
        Color = color;
    }

    public void AddPosition(Vector3 position, Vector3 direction)
    {
        positions.Add(position);

        CurrentPosition = position;
        CurrentDirection = direction;

        ignoreTransformsUntilNextStep = null;
        ignoreEntityUntilNextStep = null;
    }

    public LaserThread AddThread(Vector3 startPosition, Vector3 startDirection, float power, FlexibleColor color, Entity entityToBeIgnored = null, Transform[] transformsToBeIgnored = null)
    {
        return AddThread(new LaserThread(startPosition, startDirection, power, color, UsedEntities), entityToBeIgnored, transformsToBeIgnored);
    }

    public LaserThread AddThread(LaserThread childThread, Entity entityToBeIgnored = null, Transform[] transformsToBeIgnored = null)
    {
        if (entityToBeIgnored != null)
            childThread.SetEntityToIgnoreForNextStep(entityToBeIgnored);

        if (transformsToBeIgnored != null)
            foreach (var transformToBeIgnored in transformsToBeIgnored)
                childThread.AddToIgnoreListForNextStep(transformToBeIgnored);

        nextThreads.Add(childThread);
        return childThread;
    }

    // Checks if the transform or its entity should be ignored for the next calculation step
    public bool CheckIfShouldBeIgnored(Transform transform, Entity entity)
    {
        if (ignoreTransformsUntilNextStep != null)
        {
            if (ignoreTransformsUntilNextStep.Contains(transform))
                return true;
        }

        if (ignoreEntityUntilNextStep != null)
        {
            if (ignoreEntityUntilNextStep == entity)
                return true;
        }

        return false;
    }

    public void AddToIgnoreListForNextStep(Transform transform)
    {
        if (ignoreTransformsUntilNextStep == null)
            ignoreTransformsUntilNextStep = new List<Transform>();

        ignoreTransformsUntilNextStep.Add(transform);
    }

    public void SetEntityToIgnoreForNextStep(Entity entity)
    {
        ignoreEntityUntilNextStep = entity;
    }

    public bool IsEntityAlreadyUsed(LaserReceiverMonoBehaviour entity)
    {
        return usedEntities.Contains(entity);
    }

    // The visual thickness of the thread
    public float Thickness { get { return Power * 0.1f; } }
}
