﻿/**
 * Copyright (c) 2013 Tobias Müller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using System;
using System.Collections.Generic;
using UnityEngine;

/*
 * A laser emitter emits one or multiple laser threads and is concerned with the drawing and the
 * resoltion of the laser threads it emits. Its never used on its own, but with an entity which
 * gives it on startup the information it needs: How many lasers, where do they start and in which
 * direction are they rotated, which color and how much power do they have.
 */
public class LaserEmitter : MonoBehaviour, ILaserEmitter
{
    [SerializeField] private LineRenderer lineRendererPrefab;

    private GameManager gameManager;
    private List<LineRenderer> lineRendererList;

    // This class contains the data for a single emitted laser.
    public class LaserData
    {
        public Vector3 LocalPosition;
        public Vector3 LocalDirection;
        public float Power;
        public FlexibleColor Color;

        public LaserData() : this(Vector3.zero, new Vector3(0, 0, 1))
        {
        }

        public LaserData(Vector3 localPosition, Vector3 localDirection, float power = 1, FlexibleColor color = FlexibleColor.ABC)
        {
            this.LocalPosition = localPosition;
            this.LocalDirection = localDirection;
            this.Power = power;
            this.Color = color;
        }
    }

    public List<LaserData> LaserDataList { get; private set; }

    void Awake()
    {
        LaserDataList = new List<LaserData>();

        gameManager = GameManager.FindInstance();
        lineRendererList = new List<LineRenderer>();
    }

    /*
     * This is called every frame to update the laser and process the entities it hits - for example
     * being mirrored or doing damage to an enemy.
     */
    public void EmitLaser()
    {
        if (LaserDataList.Count == 0)
            return;

        // Create a laser node with neither position nor direction as a parent for all laser threads
        var rootLaserThread = new LaserThread(Vector3.zero, Vector3.zero, 0, FlexibleColor.ABC, null);

        // Add the laser threads. This has to be done every time because the threads will be filled with data.
        foreach (var laserData in LaserDataList)
        {
            var laserThread = new LaserThread(transform.TransformPoint(laserData.LocalPosition), transform.TransformDirection(laserData.LocalDirection), laserData.Power, laserData.Color, null);
            rootLaserThread.AddThread(laserThread);
        }

        // Calculates the laser threads and draws them
        UpdateLaserThread(rootLaserThread);
    }

    public void UpdateLaserThread(LaserThread laserThread)
    {
        // The hard limit is the maximum number a thread can be mirrored to avoid endless loops
        int hardLimit = 100;

        CalculateLaserThread(laserThread, ref hardLimit);
        DrawLaserThread(laserThread);
    }

    private void CalculateLaserThread(LaserThread laserThread, ref int hardLimit)
    {
        while (true)
        {
            // If we mirrored a certain amount of times, we just assume there's a loop going on
            if (--hardLimit < 0)
            {
                if (hardLimit == -1)
                    Debug.LogError("Reached hard limit for calculating laser thread; aborting.");

                return;
            }

            var hits = Physics.RaycastAll(laserThread.CurrentPosition, laserThread.CurrentDirection);

            // No hit? We'll just project into eternity then
            if (hits.Length == 0)
            {
                laserThread.AddPosition(laserThread.CurrentPosition + laserThread.CurrentDirection * 100, Vector3.zero);
                break;
            }

            // Sort the hits by distance
            Array.Sort(hits, (hit1, hit2) => hit1.distance.CompareTo(hit2.distance));

            // Go through hits and give them the chance to react. If they don't, we take the next one.
            foreach (var hit in hits)
            {
                Transform nextTransformToBeTested = hit.transform;
                Entity hitEntity = null;

                // Go up in the object hierarchy to find a game object which has the Entity component
                while ((hitEntity == null) && (nextTransformToBeTested != null))
                {
                    hitEntity = nextTransformToBeTested.GetComponent<Entity>();
                    nextTransformToBeTested = nextTransformToBeTested.parent;
                }

                if (hitEntity == null)
                    continue;

                if (laserThread.CheckIfShouldBeIgnored(hit.transform, hitEntity))
                    continue;

                // Process the hit. If this returns true, it was processed, if not, it was ignored and we go on.
                if (hitEntity.LaserHit(laserThread, hit))
                    break;
            }

            // Break if the laser thread should not go on, i.e. has no direction
            if (laserThread.CurrentDirection.sqrMagnitude == 0)
                break;
        }

        // Calculate child threads
        foreach (var nextThread in laserThread.NextThreads)
        {
            CalculateLaserThread(nextThread, ref hardLimit);
        }
    }

    private void DrawLaserThread(LaserThread laserThread)
    {
        int rendererIndex = 0;
        DrawSingleLaserThread(laserThread, ref rendererIndex);
        HideUnusedLineRenderers(rendererIndex);
    }

    private void DrawSingleLaserThread(LaserThread laserThread, ref int startRendererIndex)
    {
        var lineRenderer = GetOrCreateLineRenderer(startRendererIndex++);

        var color = gameManager.GetColor(laserThread.Color);
        var thickness = laserThread.Thickness * gameManager.LaserRenderStrengthMultiplier;

        //lineRenderer.material.SetColor("_Emission", color);
        lineRenderer.material.color = color;
        lineRenderer.SetColors(color, color);
        lineRenderer.SetWidth(thickness, thickness);
        lineRenderer.SetVertexCount(laserThread.Positions.Count);

        lineRenderer.transform.position = laserThread.CurrentPosition;
        lineRenderer.transform.LookAt(laserThread.Positions[0]);

        var positionIndex = 0;
        foreach (var position in laserThread.Positions)
        {
            lineRenderer.SetPosition(positionIndex++, lineRenderer.transform.InverseTransformPoint(position));
        }

        foreach (var nextThread in laserThread.NextThreads)
        {
            DrawSingleLaserThread(nextThread, ref startRendererIndex);
        }

        var lineRendererParticleSystem = lineRenderer.particleSystem;

        var showHitEmitter = laserThread.NextThreads.Count == 0;
        if (showHitEmitter)
        {
            lineRendererParticleSystem.startColor = color;

            if (!lineRendererParticleSystem.isPlaying)
                lineRendererParticleSystem.Play();
        }
        else
        {
            if (lineRendererParticleSystem.isPlaying)
                lineRendererParticleSystem.Stop();
        }
    }

    private LineRenderer GetOrCreateLineRenderer(int index)
    {
        // Create renderers up to the index if we don't have enough
        while (lineRendererList.Count <= index)
        {
            var lineRendererGameObject = (GameObject) Instantiate(lineRendererPrefab.gameObject);
            lineRendererGameObject.name = "Laser Thread Renderer #" + lineRendererList.Count;
            lineRendererGameObject.transform.parent = transform;
            lineRendererGameObject.transform.localPosition = Vector3.zero;
            lineRendererGameObject.transform.localRotation = Quaternion.identity;
            lineRendererGameObject.transform.localScale = Vector3.one;

            var lineRenderer = lineRendererGameObject.GetComponent<LineRenderer>();

            lineRendererList.Add(lineRenderer);
        }

        var currentRenderer = lineRendererList[index];
        currentRenderer.enabled = true;

        return currentRenderer;
    }

    /*
     * We only create line renderers and never remove them because creating/destroying is
     * quite expensive. We have to hide unused renderers though.
     */
    private void HideUnusedLineRenderers(int firstUnusedIndex)
    {
        for (int i = firstUnusedIndex; i < lineRendererList.Count; i++)
        {
            var lineRenderer = lineRendererList[i];

            lineRenderer.enabled = false;
            lineRenderer.particleSystem.Stop();
        }
    }

    public void Clear()
    {
        LaserDataList.Clear();
        HideUnusedLineRenderers(0);
    }
}
