/**
 * Copyright (c) 2013 Tobias Mller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using UnityEngine;

/*
 * The EnemyEntity controls a single enemy. It moves it towards the base, processes
 * laser hits and it hitting the base, and concerns itself with the various visual
 * effects attached to an enemy.
 */
public class EnemyEntity : LaserReceiverMonoBehaviour
{
    [SerializeField] private Texture texture;
    [SerializeField] private float totalFadeOutTime;

    [SerializeField] private float floatingPhase = 1;
    [SerializeField] private float floatingHeight = 0.5f;

    [SerializeField] private float damageColorMatching = 1f;
    [SerializeField] private float damageColorNotMatching = 0.5f;

    public BaseEntity Target { get; set; }
    public GameObject VisualPartContainer { get; set; }
    public float Speed { get; set; }
    public float Health { get; set; }
    public float MaxHealth { get; set; }
    public float ParticleSystemSize { get; set; }
    public FlexibleColor EnemyFlexibleColor { get; set; }
    public bool DebugMode { get; set; }
    
    private GameManager gameManager;
    private EnemyWaveQueue enemyWaveQueue;

    private MeshRenderer meshRenderer;
    private new ParticleSystem particleSystem;

    private Shader shaderOpaque;
    private Shader shaderTransparent;

    private bool dead;
    private float fadingInStartTime;
    private bool stillFadingIn;
    private bool stillFadingOut;
    private float fadeOutTime;

    private float _meshAlpha;

    private Color normalColor, fadeOutColor;

    private float randomSpawnValue;

    void Awake()
    {
        gameManager = GameManager.FindInstance();
        fadingInStartTime = gameManager.PlayState.ElapsedGameTime;

        stillFadingIn = true;
        _meshAlpha = -1;

        randomSpawnValue = Random.Range(0f, 1f);
    }

	void Start()
	{
	    var unityColor = gameManager.GetColor(EnemyFlexibleColor);

	    particleSystem = GetComponentInChildren<ParticleSystem>();
        if (particleSystem != null)
        {
	        var playFieldScale = gameManager.PlayField.Scale;
	        particleSystem.startSpeed *= playFieldScale;
            particleSystem.startLifetime *= playFieldScale;

            UpdateParticleSystemSize(1);

            particleSystem.startColor = unityColor;
            particleSystem.Play();
        }

        normalColor = unityColor;
        fadeOutColor = new Color(normalColor.r, normalColor.g, normalColor.b, 0f);

        // Find default and fade-in/out shaders
	    shaderOpaque = Shader.Find("Diffuse");
        shaderTransparent = Shader.Find("Transparent/Diffuse");

        var material = new Material(shaderTransparent);
	    material.color = new Color(normalColor.r, normalColor.g, normalColor.b, 0);
        material.mainTexture = texture;

        meshRenderer = GetComponentInChildren<MeshRenderer>();
	    meshRenderer.material = material;
	    _meshAlpha = 0;

	    TurnToTarget();
	}

    private float FadeInPercent
    {
        get { return Mathf.Clamp01((gameManager.PlayState.ElapsedGameTime - fadingInStartTime) / gameManager.TimeToShowNextWaveIndicator); }
    }

    void Update()
    {
        if (stillFadingOut)
        {
            fadeOutTime += Time.deltaTime;

            float percent = Mathf.Clamp01(fadeOutTime / totalFadeOutTime);

            if (percent == 1)
            {
                stillFadingOut = false;
            }

            MeshAlpha = 1 - percent;
        }

        if (dead)
        {
            if ((particleSystem != null) && (!particleSystem.isPlaying) && (!stillFadingOut))
            {
                Destroy(gameObject);
            }

            return;
        }

        if (floatingHeight > 0)
        {
            VisualPartContainer.transform.localPosition = new Vector3(0, -floatingHeight + Mathf.Cos((Time.time + randomSpawnValue) * floatingPhase) * floatingHeight * 0.5f, 0);
        }

        if (gameManager.GameStopped)
            return;

        if (stillFadingIn)
        {
            float percent = FadeInPercent;

            if (percent == 1)
            {
                stillFadingIn = false;
            }

            MeshAlpha = percent;

            return;
        }
        
        var direction = Target.transform.position - transform.position;
	    
        float targetDistance = direction.magnitude;
        direction.Normalize();

	    float moveDistance = Speed * gameManager.PlayField.Scale * Time.deltaTime;
        if (moveDistance > targetDistance)
        {
            Target.HitByEnemy(this);
            Die(false);
            return;
        }

        transform.Translate(direction * moveDistance, Space.World);
        TurnToTarget();
    }

    protected float MeshAlpha
    {
        get { return _meshAlpha; }
        set
        {
            value = Mathf.Clamp01(value);
            if (value == _meshAlpha)
                return;

            // If it was 1 and now isn't, we need to switch to transparent
            if (_meshAlpha == 1f)
            {
                meshRenderer.material.shader = shaderTransparent;
            }
            // If it wasn't 1 and now is, we need to switch to opaque
            else if (value == 1f)
            {
                meshRenderer.material.shader = shaderOpaque;
            }

            meshRenderer.material.color = new Color(normalColor.r, normalColor.g, normalColor.b, value);

            _meshAlpha = value;
        }
    }

    private void TurnToTarget()
    {
        transform.LookAt(Target.transform.position, Target.transform.up);
    }

    private void Die(bool wasKilled)
    {
        gameManager.PlayField.RemoveEntity(GetComponent<Entity>());

        if (particleSystem != null)
        {
            particleSystem.loop = false;
            particleSystem.emissionRate = 0;
        }

        dead = true;

        if (totalFadeOutTime > 0)
        {
            stillFadingOut = true;
            fadeOutTime = 0;
        }
        else
        {
            Destroy(meshRenderer);
        }

        if ((!stillFadingOut) && (particleSystem == null))
        {
            Destroy(gameObject);
        }

        gameManager.PlayField.EnemyDied(this, wasKilled);
    }

    public override bool LaserHit(LaserThread laserThread, RaycastHit hit)
    {
        if (dead)
            return false;

        if (stillFadingIn)
            return false;

        laserThread.AddPosition(hit.point, Vector3.zero);

        float power = laserThread.Power;

        var laserColor = laserThread.Color;

        // Find out how many components fit, how many are wrong etc.
        //int neededColorComponents = GetAmountOfColorComponents(FlexibleColor);
        int amountOfMatchingPresentColors = GetAmountOfMatchingColorComponents(EnemyFlexibleColor, laserColor, true);
        //int amountOfTotalLaserComponents = GetAmountOfColorComponents(laserColor);

        //float neededColorComponentsPresentPercentage = (float) amountOfMatchingPresentColors / neededColorComponents;
        //int amountOfDisturbingColors = amountOfTotalLaserComponents - amountOfMatchingPresentColors;

        float powerPercentage = 0;

        bool laserColorIsEnemyColor = laserColor == EnemyFlexibleColor;

        // Do any color components match at all?
        if (amountOfMatchingPresentColors > 0)
        {
            //powerPercentage = neededColorComponentsPresentPercentage - 0.1f * amountOfDisturbingColors;

            powerPercentage = laserColorIsEnemyColor
                                ? damageColorMatching
                                : damageColorNotMatching;

            power *= powerPercentage;

            if (power > 0)
            {
                Health -= power * Time.deltaTime;
                if (Health <= 0)
                {
                    Health = 0;
                    Die(true);
                }

                gameManager.PlayField.EnemyHitByLaser(power);
            }
        }

        if (DebugMode)
        {
            Debug.Log("powerPercentage: " + powerPercentage + "\n");
                      //"neededColorComponents: " + neededColorComponents + "\n" +
                      //"amountOfMatchingPresentColors: " + amountOfMatchingPresentColors + "\n" +
                      //"amountOfTotalLaserComponents: " + amountOfTotalLaserComponents + "\n" +
                      //"neededColorComponentsPresentPercentage: " + neededColorComponentsPresentPercentage + "\n" +
                      //"amountOfDisturbingColors: " + amountOfDisturbingColors
        }

        float percent = Health / MaxHealth;

        // Fade out particle system when damage is taken
        if (particleSystem != null)
        {
            particleSystem.startColor = Color.Lerp(normalColor, fadeOutColor, 1 - percent);
            UpdateParticleSystemSize(percent);
        }

        return true;
    }

    private void UpdateParticleSystemSize(float healthPercent)
    {
        particleSystem.startSize = ParticleSystemSize * healthPercent * gameManager.PlayField.Scale;
    }

    /*
     * Returns the amount of color components in a FlexibleColor
     */
    private static int GetAmountOfColorComponents(FlexibleColor color)
    {
        return (color.ContainsColor(FlexibleColor.A) ? 1 : 0) +
               (color.ContainsColor(FlexibleColor.B) ? 1 : 0) +
               (color.ContainsColor(FlexibleColor.C) ? 1 : 0);
    }

    /*
     * Returns the amount of color components which are set to [componentShouldHaveThisValue] in BOTH color1 and color2
     */
    private static int GetAmountOfMatchingColorComponents(FlexibleColor color1, FlexibleColor color2, bool componentShouldHaveThisValue)
    {
        return (IsColorSameAndValue(color1, color2, FlexibleColor.A, componentShouldHaveThisValue) ? 1 : 0) +
               (IsColorSameAndValue(color1, color2, FlexibleColor.B, componentShouldHaveThisValue) ? 1 : 0) +
               (IsColorSameAndValue(color1, color2, FlexibleColor.C, componentShouldHaveThisValue) ? 1 : 0);
    }

    /*
     * Checks if both colors have the a color component set to [componentShouldHaveThisValue]
     */
    private static bool IsColorSameAndValue(FlexibleColor color1, FlexibleColor color2, FlexibleColor component, bool componentShouldHaveThisValue)
    {
        var color1Contains = color1.ContainsColor(component);
        var color2Contains = color2.ContainsColor(component);

        return (color1Contains == color2Contains) && (color1Contains == componentShouldHaveThisValue);
    }
}
