/**
 * Copyright (c) 2013 Tobias Mller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using System;
using System.Collections.Generic;
using UnityEngine;

/**
 * The enemy factory holds several methods to spawn new enemies on the current playField. 
 * After spawning them, the EnemyEntity-Class takes over control over the spawned enemy game object.
 */
public class EnemyFactory : MonoBehaviour
{
    [SerializeField] private EnemyEntity enemyBase;
    [SerializeField] private GameObject[] enemySizePrefabs;
    [SerializeField] private GameObject[] enemySpeedPrefabs;

    [SerializeField] private float[] healthLevels;
    [SerializeField] private float[] speedLevels;
    [SerializeField] private float[] speedParticleSystemSizeLevels;

    [SerializeField] private bool debugMode;

    private GameManager gameManager;

	void Start()
	{
	    gameManager = GameManager.FindInstance();
	}
	
	void Update()
    {
	}

    public EnemyEntity Spawn(Vector3 playFieldPosition, int speedIndex, int healthIndex, FlexibleColor color)
    {
        // Create enemy object
        var enemyObject = (GameObject) Instantiate(enemyBase.gameObject);

        var visualPartContainer = new GameObject("VisualPartContainer");
        visualPartContainer.transform.parent = enemyObject.transform;
        visualPartContainer.transform.localPosition = Vector3.zero;
        visualPartContainer.transform.localRotation = Quaternion.identity;
        visualPartContainer.transform.localScale = Vector3.one;

        var sizeObject = (GameObject) Instantiate(enemySizePrefabs[healthIndex].gameObject, enemyObject.transform.position, enemyObject.transform.rotation);
        sizeObject.transform.parent = visualPartContainer.transform;

        var speedObject = (GameObject) Instantiate(enemySpeedPrefabs[speedIndex].gameObject, enemyObject.transform.position, enemyObject.transform.rotation);
        speedObject.transform.parent = visualPartContainer.transform;

        // Set values
        var enemyEntity = enemyObject.GetComponent<EnemyEntity>();

        var speed = speedLevels[speedIndex];
        var health = healthLevels[healthIndex];
        var particleSystemSize = speedParticleSystemSizeLevels[healthIndex];

        enemyEntity.Target = gameManager.PlayField.BaseEntity;
        enemyEntity.VisualPartContainer = visualPartContainer;
        enemyEntity.Speed = speed;
        enemyEntity.Health = health;
        enemyEntity.MaxHealth = health;
        enemyEntity.ParticleSystemSize = particleSystemSize;
        enemyEntity.EnemyFlexibleColor = color;
        enemyEntity.DebugMode = debugMode;

        gameManager.PlayField.AddEntity(enemyObject.GetComponent<Entity>());

        enemyEntity.transform.localPosition = playFieldPosition;

        return enemyEntity;
    }

    public Vector3 GetRandomPosition(float distance)
    {
        Vector3 localPosition = new Vector3();

        Vector2 playFieldSize = gameManager.PlayField.Size;

        bool axisX = UnityEngine.Random.value >= 0.5f;
        int direction = (UnityEngine.Random.value >= 0.5f) ? 1 : -1;
        if (axisX)
        {
            localPosition.x = UnityEngine.Random.Range(-playFieldSize.x - distance, playFieldSize.x + distance);
            localPosition.z = direction * (playFieldSize.y + distance);
        }
        else
        {
            localPosition.x = direction * (playFieldSize.x + distance);
            localPosition.z = UnityEngine.Random.Range(-playFieldSize.y - distance, playFieldSize.y + distance);
        }

        return localPosition;
    }

    /**
     * Spawns a swarm of enemies at the given distance within the swarm radius.
     * If center position is not vector3.zero, it uses this position as the center of the swarm.
     * If center position is vector3.zero, it randomly generates the center position of the swarm.
     */
    public List<EnemyEntity> SpawnSwarm(int numberOfEntities, float distance, FlexibleColor[] flexibleColors, int speedIndex, int healthIndex, float swarmRadius, Vector3 centerPosition)
    {
        var position = Vector3.zero;
        var offset = Vector3.zero;
        var enemyEntities = new List<EnemyEntity>();

        distance += swarmRadius;

        if(centerPosition == Vector3.zero)
        {
            centerPosition = GetRandomPosition(distance);
        }

        for (int i = 0; i < numberOfEntities; i++)
        {
            if (i == 0)
            {
                position = centerPosition;
            }
            else
            {
                // Chose randomly deviating positions from the center of the swarm.
                offset.x = UnityEngine.Random.Range(-swarmRadius, swarmRadius);
                offset.z = UnityEngine.Random.Range(-swarmRadius, swarmRadius);
            }

            // Chose random color of the given flexibleColors array.
            var enemyColor = flexibleColors[UnityEngine.Random.Range(0, flexibleColors.Length)];
            
            // Finally Spawn the enemy.
            enemyEntities.Add(Spawn(position + offset, speedIndex, healthIndex, enemyColor));
        }

        return enemyEntities;
    }

    /**
     * Spawns enemies randomly scattered around the playfield on the given distance.
     * If positions is not null, these positions are used instead of random positions.
     */
    public List<EnemyEntity> SpawnScattered(int numberOfEntities, float distance, FlexibleColor[] flexibleColors, int speedIndex, int healthIndex, List<Vector3> positions=null)
    {
        var enemyEntities = new List<EnemyEntity>();

        if (positions == null)
        {
            positions = new List<Vector3>();
            for (int i = 0; i < numberOfEntities; i++)
            {
                positions.Add(GetRandomPosition(distance));
            }
        }

        for (int i = 0; i < numberOfEntities; i++)
        {
            var position = positions[i];

            // Chose random color of the given flexibleColors array.
            var enemyColor = flexibleColors[UnityEngine.Random.Range(0, flexibleColors.Length)];

            // Finally Spawn the enemy.
            enemyEntities.Add(Spawn(position, speedIndex, healthIndex, enemyColor));
        }

        return enemyEntities;
    }

    /**
     * Spawns a swarm of enemies at the given distance within the swarm radius.
     * If center position is not vector3.zero, it uses this position as the center of the swarm.
     * If center position is vector3.zero, it randomly generates the center position of the swarm.
     */
    public List<EnemyEntity> SpawnLine(int numberOfEntities, float distance, FlexibleColor[] flexibleColors, int speedIndex, int healthIndex, float distanceOnLine, Vector3 startPosition) 
    {
        var position = Vector3.zero;
        var offset = Vector3.zero;
        var normalizedDirection = Vector3.zero;
        var enemyEntities = new List<EnemyEntity>();
        
        if(startPosition == Vector3.zero)
        {
            startPosition = GetRandomPosition(distance);
        }

        for (int i = 0; i < numberOfEntities; i++)
        {
            if (i == 0)
            {
                position = startPosition;
                normalizedDirection = (position - gameManager.PlayField.BaseEntity.transform.localPosition);
                normalizedDirection.Normalize();
            }
            else
            {
                // Spawn enemies on a line: Each member of the line is more far away.
                position += normalizedDirection * distanceOnLine;
            }

            // Chose random color of the given flexibleColors array.
            var enemyColor = flexibleColors[UnityEngine.Random.Range(0, flexibleColors.Length)];

            // Finally Spawn the enemy.
            enemyEntities.Add(Spawn(position + offset, speedIndex, healthIndex, enemyColor));
        }

        return enemyEntities;
    }
}
