/**
 * Copyright (c) 2013 Tobias Mller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using UnityEngine;
using System.Collections;

using System.Linq;


/**
 * This class can be used in the unity editor to organize a wave queue that has to be beaten by the player.
 * Therefore, it holds a list of waves, that can be edited. It also holds some global (default) values that may be used by each wave.
 * Everytime a new game starts, it refills the EnemyWaveQueue with its pre-initializes waves, starting the new game.
 */
public class EnemySpawnerPlanned : MonoBehaviour
{
    [SerializeField] private float distance;

    /**
     * This value is used as the default distance on line for each wave using line formation.
     * This makes it easier to globally plan the waves from within the editor
     */
    [SerializeField] private float[] distanceOnLineDefault;

    /**
     * This value is used as the default swarm radius of each wave using swarm formation.
     * This makes it easier to globally plan the waves from within the editor
     */
    [SerializeField] private float[] swarmRadiusDefault;

    /**
     * This value multiplies the time gap value of each wave. 
     * This makes it easier to globally plan the waves from within the editor
     */
    [SerializeField] private float timeGapToLastWaveMultiplier;
    [SerializeField] private EnemyWaveInfoPlanned[] waves;

    private EnemyWaveQueue enemyWaveQueue;
    private GameManager gameManager;

    void Awake()
    {
        gameManager = GameManager.FindInstance();
        enemyWaveQueue = GetComponent<EnemyWaveQueue>();

        gameManager.EventGameStarted += GameStartedListener;
    }

    private void GameStartedListener()
    {
        // Initializes a new queue of enemy waves.
        enemyWaveQueue.Initialize();
        enemyWaveQueue.Distance = distance;

        float accumulatedSpawnTime = 0;

        int totalEnemyCount = 0;

        // Fills the enemy wave queue with the waves organized via Unity editor.
        // It translates the user-friendly fields of the EnemyWaveInfoPlanned
        // to the actually used values stored in EnemyWaveInfo members.
        foreach (var wavePlanned in waves)
        {
            var newWave = new EnemyWaveInfo()
            {
                SpawnTime = accumulatedSpawnTime + (wavePlanned.TimeGapToLastWave * timeGapToLastWaveMultiplier),
                NumberOfEntities = wavePlanned.NumberOfEntities,
                FlexibleColors = wavePlanned.FlexibleColors,
                SpeedIndex = wavePlanned.SpeedIndex,
                HealthIndex = wavePlanned.HealthIndex,
                WaveFormation = wavePlanned.WaveFormation,
                DistanceOnLine = (wavePlanned.DistanceOnLine == 0f) ? distanceOnLineDefault[wavePlanned.HealthIndex] : wavePlanned.DistanceOnLine,
                SwarmRadius = (wavePlanned.SwarmRadius == 0f) ? swarmRadiusDefault[wavePlanned.HealthIndex] : wavePlanned.SwarmRadius,
            };

            enemyWaveQueue.AddWaveBack(newWave);

            accumulatedSpawnTime += wavePlanned.TimeGapToLastWave;

            totalEnemyCount += newWave.NumberOfEntities;
        }

        gameManager.PlayField.TotalEnemyCount = totalEnemyCount;
    }
}
