/**
 * Copyright (c) 2013 Tobias Mller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using UnityEngine;
using System.Collections;
using System.Collections.Generic;

/**
 * The enemy wave queue stores all waves of enemies. 
 * It constantly checks the GameManagers elapsed game time to eventually spawn the next wave.
 */
public class EnemyWaveQueue : MonoBehaviour 
{

    private List<EnemyWaveInfo> enemyWaves;
    private EnemyFactory enemyFactory;
    private GameManager gameManager;

    private HashSet<int> spawnedWaveHealthIndizes; 

    /**
     * The distance to the playField on which each wave spawns
     */
    public float Distance { get; set; }

    void Awake()
    {
        enemyFactory = GetComponent<EnemyFactory>();
        gameManager = GameManager.FindInstance();

        spawnedWaveHealthIndizes = new HashSet<int>();

        Initialize();
    }

    public void Initialize()
    {
        enemyWaves = new List<EnemyWaveInfo>();
    }

    void Start()
    {
    }
	
	void Update() 
    {
        if (gameManager.GameStopped)
            return;

        // If there are waves left and it is time to spawn the next wave in the list it gets spawned.
        while ((enemyWaves.Count > 0) && (gameManager.PlayState.ElapsedGameTime > enemyWaves[0].SpawnTime - gameManager.TimeToShowNextWaveIndicator))
        {
            var wave = enemyWaves[0];
            if (wave.WaveFormation == EnemyWaveFormation.Line)
            {
                enemyFactory.SpawnLine(wave.NumberOfEntities, Distance, wave.FlexibleColors, wave.SpeedIndex, wave.HealthIndex, wave.DistanceOnLine, Vector3.zero);
            }
            else if (wave.WaveFormation == EnemyWaveFormation.Scattered)
            {
                enemyFactory.SpawnScattered(wave.NumberOfEntities, Distance, wave.FlexibleColors, wave.SpeedIndex, wave.HealthIndex);
            }
            else if (wave.WaveFormation == EnemyWaveFormation.Swarm)
            {
                enemyFactory.SpawnSwarm(wave.NumberOfEntities, Distance, wave.FlexibleColors, wave.SpeedIndex, wave.HealthIndex, wave.SwarmRadius, Vector3.zero);
            }
            else
            {
                throw new System.ArgumentException("Enemy wave formation '" + wave.WaveFormation.ToString() + "' not implemented yet!");
            }

            enemyWaves.RemoveAt(0);

            spawnedWaveHealthIndizes.Add(wave.HealthIndex);
        }

        if (spawnedWaveHealthIndizes.Count > 0)
        {
            var soundManager = SoundManager.Instance;

            foreach (var healthIndex in spawnedWaveHealthIndizes)
            {
                soundManager.PlayMonsterAppears(healthIndex);
            }
            spawnedWaveHealthIndizes.Clear();
        }
	}

    /**
     * Fills the enemy waves with one wave.
     */
    public void AddWaveBack(EnemyWaveInfo wave)
    {
        enemyWaves.Add(wave);
    }
}
