﻿/**
 * Copyright (c) 2013 Tobias Müller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using System;
using UnityEngine;
using System.Collections;

/*
 * An entity controller provides the indirection needed between a MarkerObject and an entity on the PlayField.
 * The complete chain looks like this:
 * 
 * Physical real-world marker updates a MarkerObject (via the ARManager)
 * The GameObject which has the MarkerObject also has an instance of AbstractEntityController.
 * The AbstractEntityController creates, removes or updates an entity on the PlayField.
 * 
 * Since most PlayField entities should only exist while the marker is on the playfield (and not for example when
 * the marker is found, but outside the playfield; but when the marker is nowhere to be found, the entity might
 * want to persist), this indirection is needed.
 * 
 * Furthermore, these features are easily doable with this method:
 * a) Markers might not be allowed to move a PlayField entity directly in the current state
 * b) The PlayField entity shouldn't be moved to the new marker position instantly, but slowly move towards it.
 * c) The PlayField entity should stay on the PlayField even when the marker is currently covered (and therefore not found)
 */
public abstract class AbstractEntityController : MonoBehaviour
{
    [SerializeField] private float overridingMoveSpeed = -1;
    [SerializeField] private float overridingRotateSpeed = -1;

    protected ControllableEntity controllableEntity;
    protected MarkerObject markerObject;
    protected ControlMarker controlMarker;
    protected GameManager gameManager;

    protected bool targetInitialized;
    protected Vector3 targetPosition;
    protected Quaternion targetRotation;

    protected virtual void Awake()
    {
        markerObject = GetComponent<MarkerObject>();
        controlMarker = GetComponent<ControlMarker>();
        gameManager = GameManager.FindInstance();

        targetInitialized = false;
    }

    protected virtual void Update()
    {
        if (gameManager.GameStopped)
            return;

        var wasInUse = controllableEntity.InUse;

        if (markerObject.Detected && markerObject.Legal && controlMarker.OnPlayfield && controlMarker.Acknowledged)
        {
            controllableEntity.InUse = true;

            if (!wasInUse)
            {
                gameManager.PlayField.AddEntity(controllableEntity.GetComponent<Entity>());
            }

            var positionDistance = Vector3.Distance(targetPosition, markerObject.transform.position);
            if (positionDistance >= gameManager.ControllableGameEntityMinimalPositionDistance)
            {
                //if (positionDistance > 0) Debug.Log(positionDistance);
                targetPosition = markerObject.transform.position;
                targetPosition = gameManager.PlayField.RestrictToPlayField(targetPosition);
            }

            targetRotation = markerObject.transform.rotation;

            targetInitialized = true;
        }

        var isInUse = controllableEntity.InUse;

        if (isInUse && targetInitialized)
        {
            controllableEntity.transform.position = (MoveSpeed > 0) && wasInUse
                                                        ? Vector3.MoveTowards(controllableEntity.transform.position, targetPosition, MoveSpeed * Time.deltaTime)
                                                        : targetPosition;

            controllableEntity.transform.rotation = (RotateSpeed > 0) && wasInUse
                                                        ? Quaternion.RotateTowards(controllableEntity.transform.rotation, targetRotation, RotateSpeed * Time.deltaTime)
                                                        : targetRotation;
        }
    }

    public float MoveSpeed
    {
        get
        {
            return (overridingMoveSpeed >= 0)
                       ? overridingMoveSpeed
                       : gameManager.ControllableGameEntityMoveSpeed;
        }
    }

    public float RotateSpeed
    {
        get
        {
            return (overridingRotateSpeed >= 0)
                       ? overridingRotateSpeed
                       : gameManager.ControllableGameEntityRotateSpeed;
        }
    }
}
