﻿/**
 * Copyright (c) 2013 Tobias Müller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using System;
using UnityEngine;
using System.Collections;

/*
 * A ControlMarker is attached to the GameObject of a MarkerObject. It denote that
 * this marker can be used by the player on the PlayField and concerns itself with
 * the acknowledgement of added/removed markers. Two things are interesting here:
 * 
 * - Some markers are only to be counted as "removed" when they exist through the
 *   side of the playfield. This counters the tendency to accidently cover real-world
 *   markers.
 * - The PlayField might restrict the number of ControlMarkers on the playfield to
 *   ensure a certain level of difficulty. (This is fully implemented, but while
 *   playtesting we decided not to use it in the current prototype)
 *   For that reason, any marker who wants to be but on the PlayField has to be
 *   acknowledged by it.
 */
public class ControlMarker: MonoBehaviour
{
    [SerializeField] private bool needToExitThroughSide = true;

    private GameManager gameManager;
    private MarkerObject markerObject;

    private bool _onPlayfield;

    public float timeoutCountdown;

    public bool NeedToExitThroughSide { get { return needToExitThroughSide; } }

    private void Awake()
    {
        gameManager = GameManager.FindInstance();
        markerObject = GetComponent<MarkerObject>();

        _onPlayfield = false;
        Acknowledged = false;
    }

    private void Update()
    {
        if (OnPlayfield && !Acknowledged)
        {
            RequestAcknowledgement();
        }

        // Handle time out if it's not paused or detected and on the playfield, and if we are even using the timeout
        if (gameManager.GamePaused || !OnPlayfield || markerObject.Detected || (gameManager.HiddenIngameMarkersTimeout <= 0))
            return;

        timeoutCountdown -= Time.deltaTime;

        if (timeoutCountdown <= 0)
        {
            markerObject.MakeUnrequired();
            OnPlayfield = false;
        }
    }

    private void RequestAcknowledgement()
    {
        if (!OnPlayfield || Acknowledged)
            return;

        if (gameManager.PlayField.RequestAcknowledgement())
            Acknowledged = true;
    }

    private void FreeAcknowledgement()
    {
        if (!Acknowledged)
            return;

        gameManager.PlayField.FreeAcknowledgement();
        Acknowledged = false;
    }

    public bool OnPlayfield
    {
        get { return _onPlayfield; }
        set
        {
            if (_onPlayfield == value)
                return;
            
            _onPlayfield = value;

            if (OnPlayfield && !Acknowledged)
            {
                RequestAcknowledgement();
            }

            if (!OnPlayfield && Acknowledged)
            {
                FreeAcknowledgement();
            }
        }
    }

    public bool Acknowledged { get; set; }

    public void UpdateStatus()
    {
        if (NeedToExitThroughSide)
        {
            if (gameManager.PlayField.IsOnPlayField(transform.position))
            {
                if (!markerObject.Required)
                {
                    markerObject.MakeRequired(false);
                }
            }
            else
            {
                if (markerObject.Required)
                {
                    markerObject.MakeUnrequired();
                }
            }
        }

        if (markerObject.Detected || markerObject.Required)
        {
            OnPlayfield = gameManager.PlayField.IsOnPlayField(transform.position);
        }
        else
        {
            OnPlayfield = markerObject.Required;
        }

        if (markerObject.Detected)
        {
            timeoutCountdown = gameManager.HiddenIngameMarkersTimeout;
        }
    }

    public void RemoveFromPlayfield()
    {
        markerObject.MakeUnrequired();
        OnPlayfield = false;
    }
}
