﻿/**
 * Copyright (c) 2013 Tobias Müller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using System;
using System.Collections.Generic;
using UnityEngine;

/*
 * The ingame base situated in the middle of the playfield, emitting lasers and drawing
 * in enemies - currently represented as a wizard.
 */
public class BaseEntity : LaserReceiverMonoBehaviour
{
    [SerializeField] private GameObject wizardGameObject;

    [SerializeField] private int laserEmitCount = 2;
    [SerializeField] private float laserEmitStartAngleDeg = 0f;
    [SerializeField] private float laserEmitHeight = 0.5f;
    [SerializeField] private float laserEmitDistanceFromCenter = 0.5f;
    [SerializeField] private float laserPower = 1f;
    [SerializeField] private FlexibleColor laserColor = FlexibleColor.ABC;
    [SerializeField] private bool debugContinuousLaserUpdate = false;

    [SerializeField] private int startHealth;
    [SerializeField] private HealthOrb healthOrbPrefab;
    [SerializeField] private Transform healthOrbContainer;
    [SerializeField] private float healthOrbDistance;
    [SerializeField] private float healthOrbRotationSpeed;
    [SerializeField] private float healthOrbFloatDistance;
    [SerializeField] private float healthOrbFloatSpeed;
    [SerializeField] private float healthOrbFloatAngleMultiplier;

    [SerializeField] private float fadingOutTimeOnDeath = 2;

    public float HealthOrbDistance { get { return healthOrbDistance; } }
    public float HealthOrbRotationSpeed { get { return healthOrbRotationSpeed; } }
    public float HealthOrbFloatDistance { get { return healthOrbFloatDistance; } }
    public float HealthOrbFloatSpeed { get { return healthOrbFloatSpeed; } }
    public float HealthOrbFloatAngleMultiplier { get { return healthOrbFloatAngleMultiplier; } }

    private LaserEmitter laserEmitter;
    private List<HealthOrb> healthOrbs;

    private Color wizardColor;
    private bool fadingOut;
    private float fadeOutCountdown;

    public int StartHealth { get { return startHealth; } }
    public int Health { get; private set; }

    private int nextDyingOrb;

    void Awake()
    {
        laserEmitter = GetComponent<LaserEmitter>();

        Health = startHealth;
        //HitByEnemiesCount = 0;

        wizardColor = wizardGameObject.renderer.material.color;

        CreateHealthOrbs(startHealth);

        if (!Debug.isDebugBuild)
        {
            debugContinuousLaserUpdate = false;
        }
    }

    /*
     * Creates health orbs that float around the wizards head to denote how much health he has left
     */
    private void CreateHealthOrbs(int count)
    {
        nextDyingOrb = 0;

        healthOrbs = new List<HealthOrb>(count);

        var angle = Mathf.PI * 2 / count;
        for (var i = 0; i < count; i++)
        {
            var newOrbObject = (GameObject) Instantiate(healthOrbPrefab.gameObject);
            newOrbObject.transform.parent = healthOrbContainer;

            var newOrb = newOrbObject.GetComponent<HealthOrb>();
            newOrb.BaseEntity = this;
            newOrb.BaseAngle = angle * i;
            newOrb.UpdatePosition();

            healthOrbs.Add(newOrb);
        }
    }

    void Start()
    {
        UpdateLaserData();
    }

    /*
     * Pushes the laser information to the LaserEmitter component
     */
    private void UpdateLaserData()
    {
        laserEmitter.LaserDataList.Clear();

        if (laserEmitCount == 0)
            return;

        float angle = laserEmitStartAngleDeg * Mathf.Deg2Rad;
        float addAngle = Mathf.PI * 2 / laserEmitCount;

        for (int i = 0; i < laserEmitCount; i++)
        {
            float directionX = Mathf.Cos(angle);
            float directionZ = Mathf.Sin(angle);

            var laserData = new LaserEmitter.LaserData();
            laserData.LocalPosition = new Vector3(directionX * laserEmitDistanceFromCenter, laserEmitHeight, directionZ * laserEmitDistanceFromCenter);
            laserData.LocalDirection = new Vector3(directionX, 0, directionZ);
            laserData.Power = laserPower;
            laserData.Color = laserColor;

            laserEmitter.LaserDataList.Add(laserData);

            angle += addAngle;
        }
    }

    void Update()
    {
        if (debugContinuousLaserUpdate)
            UpdateLaserData();

        if (fadingOut)
        {
            fadeOutCountdown -= Time.deltaTime;
            float percentVisible = Mathf.Clamp01(fadeOutCountdown / fadingOutTimeOnDeath);

            wizardColor.a = percentVisible;
            wizardGameObject.renderer.material.color = wizardColor;

            if (percentVisible == 0)
                fadingOut = false;
        }
    }

    public override bool LaserHit(LaserThread laserThread, RaycastHit hit)
    {
        // The base ignores the laser
        return false;
    }

    public void HitByEnemy(EnemyEntity enemyEntity)
    {
        if (Health <= 0)
            return;

        RemoveHealthOrb();

        Health--;

        SoundManager.Instance.PlayMonsterEatsWizard();

        if (Health == 0)
        {
            GameManager.FindInstance().PlayField.GameOver();
        }
    }

    private void RemoveHealthOrb()
    {
        if (Health <= 0)
            return;

        // Find the next non-dead health orb
        while (healthOrbs[nextDyingOrb].IsDead)
        {
            nextDyingOrb = (nextDyingOrb + 1) % healthOrbs.Count;
        }

        // Fade that one out
        healthOrbs[nextDyingOrb].Die();

        // Jump to the next one in a certain pattern to make it look more interesting
        nextDyingOrb = (nextDyingOrb + Math.Max(Health / 3, 1)) % healthOrbs.Count;
    }

    public float LaserPower
    {
        get { return laserPower; }

        set
        {
            if (laserPower == value)
                return;

            laserPower = value;
            UpdateLaserData();
        }
    }

    public void Die()
    {
        // Let the emitters vanish
        laserEmitter.Clear();

        // Create a material to fade the wizard out
        var transparentWizardMaterial = new Material(wizardGameObject.renderer.material);
        transparentWizardMaterial.shader = Shader.Find("Transparent/Diffuse");
        wizardGameObject.renderer.material = transparentWizardMaterial;

        fadingOut = true;
        fadeOutCountdown = fadingOutTimeOnDeath;
    }
}
