﻿/**
 * Copyright (c) 2013 Tobias Müller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using System;
using System.Collections.Generic;
using UnityEngine;
using System.Collections;

/*
 * A MirrorEntity takes a LaserThread and blocks, ignores or mirrors it, based on how
 * often the LaserThread already hit the mirror and the mirrors properties.
 * The first time the laser ia always mirrored (if it hit the reflective surface part, which
 * currently encompasses the whole mirror for playing reasons), the second and further
 * times are decided by the SecondTimeLaserHitsBehaviour.
 */
class MirrorEntity : LaserReceiverMonoBehaviour
{
    private enum SecondTimeLaserHitsBehaviour
    {
        ReflectLaser,
        IgnoreLaser,
        BlockLaser
    }

    [SerializeField] private SecondTimeLaserHitsBehaviour secondTimeLaserHitsBehaviour;

    public override bool LaserHit(LaserThread laserThread, RaycastHit hit)
    {
        if (hit.collider.name == "ColliderFrame")
        {
            return LaserHitFrame(laserThread, hit);
        }

        return LaserHitMirror(laserThread, hit);
    }

    private bool LaserHitMirror(LaserThread laserThread, RaycastHit hit)
    {
        if (secondTimeLaserHitsBehaviour == SecondTimeLaserHitsBehaviour.IgnoreLaser)
        {
            if (laserThread.IsEntityAlreadyUsed(this))
                return false;
        }

        var inDirection = laserThread.CurrentDirection;
        var inNormal = hit.normal;

        laserThread.AddPosition(hit.point, Vector3.zero);

        if (secondTimeLaserHitsBehaviour == SecondTimeLaserHitsBehaviour.BlockLaser)
        {
            if (laserThread.IsEntityAlreadyUsed(this))
                return true;
        }

        var usedEntities = new List<LaserReceiverMonoBehaviour>(laserThread.UsedEntities);
        usedEntities.Add(this);

        laserThread.AddThread(new LaserThread(hit.point, Vector3.Reflect(inDirection, inNormal), laserThread.Power, laserThread.Color, usedEntities));

        return true;
    }

    private bool LaserHitFrame(LaserThread laserThread, RaycastHit hit)
    {
        laserThread.AddPosition(hit.point, Vector3.zero);
        return true;
    }
}
