﻿/**
 * Copyright (c) 2013 Tobias Müller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using System;
using System.Collections.Generic;
using UnityEngine;
using System.Collections;

/*
 * The PrismEntity splits an incoming laser thread in its color components.
 */
class PrismEntity : LaserReceiverMonoBehaviour
{
    [SerializeField] private MeshRenderer meshRenderer;
    [SerializeField] private Material unusedMaterial;
    [SerializeField] private Material usedMaterial;

    [SerializeField] private Transform center;
    [SerializeField] private bool alwaysUseSameAngles = true;
    [SerializeField] private bool ignorePrimaryColorRays = true;

    public void OnMessageBeforeEmitting()
    {
        if (unusedMaterial != null)
            meshRenderer.material = unusedMaterial;
    }

    public override bool LaserHit(LaserThread laserThread, RaycastHit hit)
    {
        //Debug.Log("Hit prism");
        var thisEntity = GetComponent<Entity>();

        //var inDirection = laserThread.CurrentDirection;
        var inColor = laserThread.Color;

        bool containsA = inColor.ContainsColor(FlexibleColor.A);
        bool containsB = inColor.ContainsColor(FlexibleColor.B);
        bool containsC = inColor.ContainsColor(FlexibleColor.C);

        int colorAmount = (containsA ? 1 : 0) + (containsB ? 1 : 0) + (containsC ? 1 : 0);

        if ((colorAmount <= 1) && ignorePrimaryColorRays)
            return false;

        laserThread.AddPosition(hit.point, Vector3.zero);

        var centerPosition = center.TransformPoint(0, center.InverseTransformPoint(hit.point).y, 0);

        laserThread.AddPosition(centerPosition, Vector3.zero);

        var directionToCenter = centerPosition - hit.point;

        var startDirectionLocal = transform.InverseTransformDirection(directionToCenter);
        var angle = Mathf.Atan2(startDirectionLocal.z, startDirectionLocal.x);

        int rayAmount;
        float addAngle = 45 * Mathf.Deg2Rad;

        if (alwaysUseSameAngles)
        {
            rayAmount = 3;
        }
        else
        {
            rayAmount = colorAmount;
            //addAngle = Mathf.PI * 2 / rayAmount;
        }

        switch (rayAmount)
        {
            case 3:
                angle -= addAngle;
                break;

            case 2:
                angle -= addAngle / 2;
                break;

        }

        if (containsA)
        {
            laserThread.AddThread(centerPosition, transform.TransformDirection(Mathf.Cos(angle), 0, Mathf.Sin(angle)), laserThread.Power, FlexibleColor.A, thisEntity);
        }

        if (alwaysUseSameAngles || containsA)
            angle += addAngle;

        if (containsB)
        {
            laserThread.AddThread(centerPosition, transform.TransformDirection(Mathf.Cos(angle), 0, Mathf.Sin(angle)), laserThread.Power, FlexibleColor.B, thisEntity);
        }

        if (alwaysUseSameAngles || containsB)
            angle += addAngle;

        if (containsC)
        {
            laserThread.AddThread(centerPosition, transform.TransformDirection(Mathf.Cos(angle), 0, Mathf.Sin(angle)), laserThread.Power, FlexibleColor.C, thisEntity);
        }

        if (usedMaterial != null)
            meshRenderer.material = usedMaterial;

        return true;
    }
}
