/**
 * Copyright (c) 2013 Tobias Mller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using System;
using System.Linq;
using UnityEngine;
using System.Collections;

/*
 * The GameManager is the heart of the game; it initializes the state machine, and its first state,
 * holds information regarding the current flow of the game and distributes information.
 * 
 * It also creates the PlayField on which the actual game happens.
 */
public class GameManager : MonoBehaviour
{
    public delegate void GameStartedDelegate();

    private ARManager arManager;
    private ARConsoleManager arConsoleManager;

    // The speed with which controlled entities interpolate to the position/rotation of their marker
    [SerializeField] private float controllableGameEntityMoveSpeed = 40;
    [SerializeField] private float controllableGameEntityRotateSpeed = 500;

    // Prevent jittering by only changing the entity if the marker changes its position at least by this value
    [SerializeField] private float controllableGameEntityMinimalPositionDistance = 0.1f;

    [SerializeField] private GameObject playfieldPrefab;
    [SerializeField] private Vector3 playfieldOffset = new Vector3(-8, 0, -6.5f);
    [SerializeField] private FlexibleColorAssignment flexibleColorAssignment;
    [SerializeField] private float laserRenderStrengthMultiplier = 2;
    [SerializeField] private float timeToShowNextWaveIndicator;

    // If a marker is suddenly hidden (might just be a hand over it), it stays ingame until it times out. 0 = no timeout
    [SerializeField] private float hiddenIngameMarkersTimeout = 5;

    [SerializeField] private int delayBeforeNewGameStarts = 6;
    [SerializeField] private string newGameStartsText = "New game starting ... ";
    
    [SerializeField] private int delayBeforeGameResumes = 3;
    [SerializeField] private string resumedGameStartsText = "Game resuming ... ";

    [SerializeField] private GUISkin defaultGuiSkin;

    [SerializeField] private bool cheatCodesActive = true;
    [SerializeField] private float cheatEnhanceLaserStrength = 2;

    // Skip right into the game once you see the origin marker
    [SerializeField] private bool debugSkipConsole;

    [SerializeField] private bool debugGameOverDeactivated;
    [SerializeField] private bool debugAllowNonConsoleMarkersInMenuState = false;

    private int timeBeforeCurrentGameStarts;
    private string currentGameStartsText;
    
    /**
     * The state the game is currently in
     */
    private State currentState;

    // All states the game can be in.
    private StateBeforeStart beforeStartState;
    private StateMenu menuState;
    private StatePlay playState;
    private StateSkipConsoleInitialization skipConsoleInitializationState;

    /**
     * The actual field the game is played on. This serves as parent for all game objects.
     */
    private PlayField playField;

    private MusicPlayer musicPlayer;

    public float ControllableGameEntityMoveSpeed { get { return controllableGameEntityMoveSpeed; } }
    public float ControllableGameEntityRotateSpeed { get { return controllableGameEntityRotateSpeed; } }
    public float ControllableGameEntityMinimalPositionDistance { get { return controllableGameEntityMinimalPositionDistance; } }
    public float LaserRenderStrengthMultiplier { get { return laserRenderStrengthMultiplier; } }

    public State CurrentState { get { return currentState; } }
    public StatePlay PlayState { get { return playState; } }
    public StateMenu MenuState { get { return menuState; } }
    public StateBeforeStart BeforeStartState { get { return beforeStartState; } }

    public PlayField PlayField { get { return playField; } }

    public event GameStartedDelegate EventGameStarted;

    public bool DebugSkipConsole { get { return debugSkipConsole; } }

    // Cancels the countdown to a new game
    private bool cancelNewGamePrelude = false;

    public bool GamePaused { get; set; }
    public bool GameOver { get; set; }
    public bool GameStopped { get { return GamePaused || GameOver; } }

    public float TimeToShowNextWaveIndicator { get { return timeToShowNextWaveIndicator; } }
    public bool DebugGameOverDeactivated { get { return Debug.isDebugBuild && debugGameOverDeactivated; } }
    public bool DebugAllowNonConsoleMarkersInMenuState { get { return Debug.isDebugBuild && debugAllowNonConsoleMarkersInMenuState; } }

    public float HiddenIngameMarkersTimeout { get { return hiddenIngameMarkersTimeout; } }

    public GUISkin DefaultGuiSkin { get { return defaultGuiSkin; } }

    public bool CheatCodesActive { get { return cheatCodesActive; } }

    public float CheatEnhanceLaserStrength { get { return cheatEnhanceLaserStrength; } }

    private void Awake()
    {
        arManager = ARManager.FindInstance();
        arManager.CallWhenInitialized(Initialize);
    }

    private void Initialize()
    {
        arConsoleManager = ARConsoleManager.FindInstance();
        musicPlayer = MusicPlayer.FindInstance();

        if (!Debug.isDebugBuild)
        {
            debugAllowNonConsoleMarkersInMenuState = false;
            debugGameOverDeactivated = false;
            debugSkipConsole = false;
        }

        // Hook into the AR Manager
	    arManager.EventBeforeUpdate += BeforeARUpdate;
        arManager.EventAfterUpdate += AfterARUpdate;
        arManager.DelegateAcknowledgeNewRecognition = AcknowledgeNewRecognition;
        arManager.DelegateAcknowledgeUpdate = AcknowledgeUpdate;
        arManager.EventDetectionStarted += DetectionStarted;
        arManager.EventDetectionUpdated += DetectionUpdated;
        arManager.EventDetectionEnded += DetectionEnded;

        arConsoleManager.ConsoleActivatedEvent += ConsoleActivatedListener;

        // Initialize the states
        beforeStartState = new StateBeforeStart();
        beforeStartState.Initialize();

        menuState = new StateMenu();
        menuState.Initialize();

        playState = new StatePlay();
        playState.Initialize();

        skipConsoleInitializationState = new StateSkipConsoleInitialization();
        skipConsoleInitializationState.Initialize();

        if (debugSkipConsole)
        {
            SwitchToState(skipConsoleInitializationState);
        }
        else
        {
            SwitchToState(beforeStartState);
        }

        InitializeNewGame();

        playField.GameStartsText.SetActive(false);
    }

    private void ConsoleActivatedListener()
    {
        // If the console is being replaced while preluding a new game prevent the new game.
        if (timeBeforeCurrentGameStarts > 0)
        {
            cancelNewGamePrelude = true;
        }
    }

    /**
     * Preludes a new game within a specified time.
     */
    public void PreludeNewGame()
    {
        timeBeforeCurrentGameStarts = delayBeforeNewGameStarts;
        currentGameStartsText = newGameStartsText;
        PreludeGame();
    }
    
    /**
     * Preludes resuming a game within a specified time.
     */
    public void PreludeResumedGame()
    {
        timeBeforeCurrentGameStarts = delayBeforeGameResumes;
        currentGameStartsText = resumedGameStartsText;
        PreludeGame();
    }

    /**
     * Preludes a new/paused game using recursion in combination with time dependent method invocation.
     */
    private void PreludeGame()
    {
        PlayField.HowToStartText.SetActive(false);
        playField.GameStartsText.SetActive(true);
        Invoke("PreludeGameRecursive", 0);
    }

    /**
     * This resursive method calls himself every second reducing the time till the game resumes/starts by 1.
     * When the remaining time before the game resumes/starts is less than one the actual game starts/returnes.
     */
    private void PreludeGameRecursive()
    {
        if (cancelNewGamePrelude)
        {
            cancelNewGamePrelude = false;
            return;
        }
        if (timeBeforeCurrentGameStarts < 1)
        {
            playField.GameStartsText.SetActive(false);
            SwitchToState(playState);
            menuState.NewGameRequested = false;
            return;
        }
        playField.GameStartsText.GetComponent<TextMesh>().text = currentGameStartsText + timeBeforeCurrentGameStarts--;
        Invoke("PreludeGameRecursive", 1);
    }

    public void InitializeNewGame()
    {
        // Get old flip settings to correctly flip camera after restarting game.
        var flipCamera = FlipCamera.FindInstance();
        var oldFlipX = flipCamera.FlipX;
        var oldFlipY = flipCamera.FlipY;
        flipCamera.FlipX = true;
        flipCamera.FlipY = false;

        float scale = 1f;

        if (playField != null)
        {
            scale = playField.Scale;

            // Save control markers from being destroyed (they are children of the playField).
            var controlMarkers = (ControlMarker[]) FindSceneObjectsOfType(typeof (ControlMarker));
            foreach (var controlMarker in controlMarkers)
            {
                controlMarker.RemoveFromPlayfield();
            }

            // Destroy the playField and all of its children.
            Destroy(playField.gameObject);
        }

        GameOver = false;

        if (!DebugAllowNonConsoleMarkersInMenuState)
            GamePaused = true;

        CreateNewPlayField();

        if (EventGameStarted != null)
            EventGameStarted();

        playState.ResetState();

        menuState.UpdateGameStartText();

        musicPlayer.NewGameWasInitialized();

        ARConsoleManager.FindInstance().DeactivateButton(ARConsoleManager.ConsoleCommandType.RestartGame);
        
        // Correctly re-flip camera in new playfield.
        if (!oldFlipX)
        {
            flipCamera.ToggleFlipX();
            playField.FlipTextsX();
        }
        if (oldFlipY)
        {
            flipCamera.ToggleFlipY();
            playField.FlipTextsY();
        }

        playField.Scale = scale;
    }

    public void GameUpdate(float elapsedGameTime)
    {
        playField.GameUpdate(elapsedGameTime);
    }

    private void CreateNewPlayField()
    {
        playField = ((GameObject)Instantiate(playfieldPrefab)).GetComponent<PlayField>();
        playField.transform.parent = arManager.OriginMarkerObject.transform;
        playField.transform.localPosition = playfieldOffset;
        playField.transform.localRotation = Quaternion.identity;
    }

    private void BeforeARUpdate()
    {
        currentState.BeforeARUpdate();
    }

    private void AfterARUpdate()
    {
        if (CanExitTheGame)
        {
            if (Input.GetButtonDown("ExitWhileConsoleIsOpen"))
            {
                Application.Quit();
            }
        }

        currentState.AfterARUpdate();
    }

    private bool AcknowledgeNewRecognition(MarkerObject markerObject)
    {
        return currentState.AcknowledgeNewRecognition(markerObject);
    }

    private bool AcknowledgeUpdate(MarkerObject markerObject)
    {
        return currentState.AcknowledgeUpdate(markerObject);
    }

    public void DetectionStarted(MarkerObject markerObject)
    {
        currentState.DetectionStarted(markerObject);
    }

    public void DetectionUpdated(MarkerObject markerObject)
    {
        currentState.DetectionUpdated(markerObject);
    }

    public void DetectionEnded(MarkerObject markerObject)
    {
        currentState.DetectionEnded(markerObject);
    }

    /**
     * Switches from one state to another
     * notifying the old and new state about the state switch.
     */
    public void SwitchToState(State newState)
    {
        if (newState == null)
            throw new Exception("Cannot set GameManager.state to an empty state");

        if (currentState != null)
        {
            currentState.Deactivate();
        }

        currentState = newState;
        currentState.Activate();
    }

    public Color GetColor(FlexibleColor flexibleColor)
    {
        return flexibleColorAssignment.GetColor(flexibleColor);
    }

    public void OnGUI()
    {
        if (!CanExitTheGame)
            return;

        GUI.skin = defaultGuiSkin;

        string text = "Press Escape to exit the game.";

        Vector2 size = GUI.skin.label.CalcSize(new GUIContent(text));
        Rect rect = new Rect(Screen.width / 2f - size.x / 2, Screen.height - size.y, size.x, size.y);

        GUI.skin.box.alignment = TextAnchor.MiddleCenter;
        GUI.Label(rect, text);
    }

    protected bool CanExitTheGame
    {
        get { return (!Application.isWebPlayer) && ((currentState == beforeStartState) || (currentState == menuState)); }
    }

    public string GetDebugString()
    {
        string text = "Control Markers: " + playField.UsedControlMarkers + "/" + playField.AllowedControlMarkers +
            "\n" + currentState.GetDebugString();

        return text;
    }

    public static GameManager FindInstance()
    {
        return UnityHelper.FindFirstActiveInstance<GameManager>();
    }
}
