/**
 * Copyright (c) 2013 Tobias Mller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using System;
using System.Collections.Generic;
using UnityEngine;
using System.Collections;

/*
 * The PlayField is the place where the actual game happens. It creates all its basic
 * entities, keeps track of them and the ControlMarkers, tells the LaserEmitters to emit
 * and ends the game if certain win/lose conditions are met.
 */
public class PlayField : MonoBehaviour
{
    [SerializeField] private GameObject singleLaserEmitterPrefab;
    [SerializeField] private GameObject wallPrefab;
    [SerializeField] private GameObject floorPrefab;
    [SerializeField] private GameObject basePrefab;
    [SerializeField] private GameObject textPrefab;
    [SerializeField] private bool showFloor;
    [SerializeField] private float floorAlpha;
    [SerializeField] private Vector2 wallCount;

    // "Control markers" are the markers used by the player to control ingame entities.
    // "allowedControlMarkers" contains the maximum count of allowed markers. The idea
    // was that the player can only have for example 3 markers at the beginning of the
    // game and continuosly more, but later on it turned out that it's complicated
    // enough without this artificial limit.
    // If this value is < 0, we allow an infinite number of markers.
    [SerializeField] private int allowedControlMarkers;

    [SerializeField] private float scale = 1f;

    private List<ILaserEmitter> laserEmitters;
    private List<Entity> entities;

    private Vector2 size;

    private Floor floor;
    private BaseEntity baseEntity;

    // How many control markers are in use right now
    public int UsedControlMarkers { get; set; }

    public BaseEntity BaseEntity { get { return baseEntity; } }
    
    public GameObject HowToStartText { get; private set; }
    public GameObject GameStartsText { get; private set; }
    public TextMesh KilledEnemiesText { get; private set; }
    public TextMesh GameOverText { get; private set; }

    public int KilledEnemyCount { get; private set; }

    // Vanishes = Killed Enemies + Enemies that damaged us
    public int VanishedEnemyCount { get; private set; }

    // Total enemies which will be spawned
    public int TotalEnemyCount { get; set; }

    private GameManager gameManager;

    // The total laser power with which enemies were hit this frame - used for sound output
    private float enemiesHitByLaserTotalPower;

    // Cheat mode to test the game
    private bool cheated;
    private bool cheatNoGameOver;

    void Awake()
    {
        gameManager = GameManager.FindInstance();

        laserEmitters = new List<ILaserEmitter>();
        entities = new List<Entity>();

        size = CreateWalls(wallCount.x, wallCount.y);

        CreateFloor();
        CreateBaseEntity();

        //CreateSingleLaserEmitter(new Vector2(size.x - 0.5f, 0), -90);

        HowToStartText = Create3DText(textPrefab, new Vector3(0, 0, -size.y), "");
        GameStartsText = Create3DText(textPrefab, new Vector3(0, 0, -size.y), "");
        KilledEnemiesText = Create3DText(textPrefab, new Vector3(0, 3, -size.y), "").GetComponent<TextMesh>();
        GameOverText = Create3DText(textPrefab, new Vector3(0, 4, 0), "").GetComponent<TextMesh>();

        cheated = false;
        cheatNoGameOver = false;

        UsedControlMarkers = 0;

        KilledEnemyCount = 0;
        VanishedEnemyCount = 0;
        TotalEnemyCount = int.MaxValue;

        UpdateKilledEnemiesText();

        UpdateFloor();
    }

    public void GameUpdate(float elapsedGameTime)
    {
        foreach (var entity in entities)
        {
            entity.SendMessage("OnMessageIngameUpdate", SendMessageOptions.DontRequireReceiver);
        }

        enemiesHitByLaserTotalPower = 0;

        foreach (var entity in entities)
        {
            entity.SendMessage("OnMessageBeforeEmitting", SendMessageOptions.DontRequireReceiver);
        }

        foreach (var laserEmitter in laserEmitters)
        {
            laserEmitter.EmitLaser();
        }

        SoundManager.Instance.MonsterWasHitByLaserTotalPower = enemiesHitByLaserTotalPower;
    }

    public void EnemyHitByLaser(float power)
    {
        enemiesHitByLaserTotalPower += power;
    }

    private void UpdateFloor()
    {
        floor.Show = showFloor;
        floor.Alpha = floorAlpha;
    }

    public GameObject Create3DText(GameObject prefab, Vector3 position, String text)
    {
        var newObject = (GameObject) Instantiate(prefab);
        newObject.transform.parent = transform;
        newObject.transform.localPosition = position;
        newObject.transform.parent = transform;
        //newObject.transform.localScale = new Vector3(0.8f,0.8f,0.8f);
        //newObject.transform.localScale *= 0.8f;

        if (text != null)
        {
            var textMesh = newObject.GetComponent<TextMesh>();
            textMesh.text = text;
        }

        return newObject;
    }

    private void CreateFloor()
    {
        floor = ((GameObject)Instantiate(floorPrefab)).GetComponent<Floor>();
        floor.transform.parent = transform;
        floor.transform.localPosition = Vector3.zero;
        floor.transform.localScale = new Vector3(size.x * 2, 1, size.y * 2);
    }

    private void CreateBaseEntity()
    {
        baseEntity = CreateAndAddObject(basePrefab, Vector2.zero, 0).GetComponent<BaseEntity>();
    }

    public void CreateSingleLaserEmitter(Vector2 position, float angle)
    {
        CreateAndAddObject(singleLaserEmitterPrefab, position, angle);
    }

    // Creates the 4 walls limiting the playing field
    public Vector2 CreateWalls(float countX, float countY)
    {
        var length = wallPrefab.GetComponent<WallEntity>().Length;

        var lengthX = length * countX;
        var lengthY = length * countY;

        for (int i = 0; i < countX; i++)
        {
            var position = new Vector2(-lengthX / 2 + i * length, -lengthY / 2);

            CreateWall(position, 0);
            CreateWall(-position, 180);
        }

        for (int i = 0; i < countY; i++)
        {
            var position = new Vector2(-lengthX / 2, -lengthY / 2 + (i + 1) * length);

            CreateWall(position, 90);
            CreateWall(-position, 270);
        }

        return new Vector2(lengthX / 2, lengthY / 2);
    }

    public void CreateWall(Vector2 position, float angle)
    {
        CreateAndAddObject(wallPrefab, position, angle);
    }

    public GameObject CreateAndAddObject(GameObject prefab, Vector2 position, float angle)
    {
        var newObject = (GameObject) Instantiate(prefab);
        newObject.transform.parent = transform;
        newObject.transform.localPosition = new Vector3(position.x, 0, position.y);
        newObject.transform.RotateAround(newObject.transform.position, Vector3.up, angle);

        AddEntity(newObject.GetComponent<Entity>());

        var laserEmitter = newObject.GetComponent<LaserEmitter>();
        if (laserEmitter != null)
            laserEmitters.Add(laserEmitter);

        return newObject;
    }

    public void AddEntity(Entity entity)
    {
        entity.transform.parent = transform;
        entity.transform.localScale = Vector3.one;

        entities.Add(entity);
    }

    public void RemoveEntity(Entity entity)
    {
        entity.transform.parent = null;

        entities.Remove(entity);
    }

    // Takes a position and returns
    // a) the position if it is on the play field or
    // b) the nearest position on the play field if it isn't
    public Vector3 RestrictToPlayField(Vector3 targetPosition)
    {
        if (IsOnPlayField(targetPosition))
            return targetPosition;

        return floor.FloorCollider.ClosestPointOnBounds(targetPosition);
    }

    public bool IsOnPlayField(Vector3 targetPosition)
    {
        RaycastHit hitInfo;
        return floor.FloorCollider.Raycast(new Ray(targetPosition + floor.transform.up, floor.transform.up * -1), out hitInfo, 2);
    }

    public int AllowedControlMarkers
    {
        get { return allowedControlMarkers; }
        set { allowedControlMarkers = value; }
    }

    // The boundaries of the playing field
    // Note: This is the distance from the center
    public Vector2 Size
    {
        get { return size; }
    }

    // The zoom factor for the playing field and everything on it
    public float Scale
    {
        get { return scale; }
        set
        {
            scale = value;
            transform.localScale = new Vector3(scale, scale, scale);
        }
    }

    // Earlier on we introduced a feature allowing only a certain number of controlled entities
    // on the playing field. This function asks if there is place for a new one, and if there is
    // it immediately reserves it.
    public bool RequestAcknowledgement()
    {
        if ((UsedControlMarkers < AllowedControlMarkers) || (AllowedControlMarkers < 0))
        {
            UsedControlMarkers++;
            return true;
        }

        return false;
    }

    // Frees a used control marker space.
    public void FreeAcknowledgement()
    {
        UsedControlMarkers--;
    }

    public void EnemyDied(EnemyEntity enemyEntity, bool killedEnemy)
    {
        if (killedEnemy)
        {
            KilledEnemyCount++;
            UpdateKilledEnemiesText();

            SoundManager.Instance.PlayMonsterDies();
        }

        VanishedEnemyCount++;

        if (VanishedEnemyCount == TotalEnemyCount)
        {
            WonGame();
        }
    }

    private void UpdateKilledEnemiesText()
    {
        KilledEnemiesText.text = "Banished Ghosts: " + KilledEnemyCount;
    }

    public void GameOver()
    {
        if (gameManager.GameOver || gameManager.DebugGameOverDeactivated || cheatNoGameOver)
            return;

        GameOverText.text = "No magical orbs left! Game Over!";
        // "Your enemies proved to be too strong for you this time."

        gameManager.GameOver = true;

        baseEntity.Die();

        SoundManager.Instance.PlayGameOver();
        MusicPlayer.FindInstance().PlayerLost();
    }

    public void WonGame()
    {
        if (gameManager.GameOver || gameManager.DebugGameOverDeactivated)
            return;

        GameOverText.text = "Congratulations, you won!";
        // "As the last of the ghosts dies (again), you remain - victorious!"

        gameManager.GameOver = true;

        SoundManager.Instance.PlayGameWon();
        MusicPlayer.FindInstance().PlayerWon();
    }

    // The following methods flip the text entities too when the playfield is flipped

    public void FlipTextsX()
    {
        FlipTextX(GameStartsText);
        FlipTextX(HowToStartText);
        FlipTextX(KilledEnemiesText.gameObject);
        FlipTextX(GameOverText.gameObject);
    }

    private void FlipTextX(GameObject text)
    {
        var localScale = text.transform.localScale;
        text.transform.localScale = new Vector3(-localScale.x, localScale.y, localScale.z);
    }
    
    public void FlipTextsY()
    {
        FlipTextY(GameStartsText);
        FlipTextY(HowToStartText);
        FlipTextY(KilledEnemiesText.gameObject);
        FlipTextY(GameOverText.gameObject);
    }

    private void FlipTextY(GameObject text)
    {
        var localScale = text.transform.localScale;
        text.transform.localScale = new Vector3(localScale.x, -localScale.y, localScale.z);
        
        var localPosition = text.transform.localPosition;
        if (localPosition.y == 0)
        {
            text.transform.localPosition = new Vector3(localPosition.x, size.y / 7, localPosition.z);
        }
        else
        {
            text.transform.localPosition = new Vector3(localPosition.x, 0, localPosition.z);
        }
    }

    public void CheatEnhanceLaser()
    {
        baseEntity.LaserPower = gameManager.CheatEnhanceLaserStrength;
        cheated = true;
    }

    public void CheatNoGameOver()
    {
        cheatNoGameOver = true;
        cheated = true;
    }

    // Output that the cheat mode was used, if that is the case
    private void OnGUI()
    {
        if (!cheated)
            return;

        GUI.skin = null;

        string text = "Cheat mode active!";

        Vector2 size = GUI.skin.box.CalcSize(new GUIContent(text));
        Rect rect = new Rect(Screen.width / 2 - size.x / 2, Screen.height - size.y - 5, size.x, size.y);

        GUI.skin.box.alignment = TextAnchor.MiddleCenter;
        GUI.Box(rect, text);

    }
}
