/**
 * Copyright (c) 2013 Tobias Mller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using System;
using UnityEngine;
using System.Collections;

/*
 * The GameManager contains a state machine controlling the flow of the game. Each
 * state is derived from this abstract State class.
 * 
 * When the state is created at the beginning, Initialize is called.
 * 
 * States are persistent and will only be activated/deactivated.
 * When a state is activated, Deactivate() will be called on the previous one before
 * calling Activate() on the new state.
 * 
 * A state contains several event handlers which are called by the GameManager and
 * two delegates (Acknowledge*) which control in which state which markers can be
 * added or updated.
 */
public abstract class State
{
    protected ARManager arManager;
    protected GameManager gameManager;
    protected ARConsoleManager arConsoleManager;

    public virtual void Initialize()
    {
        arManager = ARManager.FindInstance();
        gameManager = GameManager.FindInstance();
        arConsoleManager = ARConsoleManager.FindInstance();
    }

    public abstract void Activate();
    public abstract void Deactivate();

    public abstract void BeforeARUpdate();
    public abstract void AfterARUpdate();
    public abstract bool AcknowledgeNewRecognition(MarkerObject markerObject);
    public abstract bool AcknowledgeUpdate(MarkerObject markerObject);
    public abstract void DetectionStarted(MarkerObject markerObject);
    public abstract void DetectionUpdated(MarkerObject markerObject);
    public abstract void DetectionEnded(MarkerObject markerObject);

    public string GetDebugString()
    {
        string text = "State: " + GetType().Name;
        return text;
    }
}
