﻿/**
 * Copyright (c) 2013 Tobias Müller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using System;
using UnityEngine;
using System.Collections;

/*
 * The menu state is active when the console is shown. It serves as the main menu and the pause menu.
 * Therefore it subsrcibes to several console command events of the ARConsoleManager class.
 */
public class StateMenu : State
{
    private FlipCamera flipCamera;
    private GameObject howToStartText;
    private bool firstActivation = true;

    public bool NewGameRequested { get; set; }

    public override void Initialize()
    {
        base.Initialize();

        flipCamera = FlipCamera.FindInstance();
        
        // Subscribe to all different kind of console events because the menu needs to implement the corresponding logic.
        arConsoleManager.ConsoleActivatedEvent += ConsoleActivatedListener;
        arConsoleManager.ConsoleDeactivatedEvent += ConsoleDeactivatedListener;
        arConsoleManager.AddConsoleCommand(ARConsoleManager.ConsoleCommandType.ToggleCalibration, ToggleCalibrationListener);
        arConsoleManager.AddConsoleCommand(ARConsoleManager.ConsoleCommandType.IncrementSize, IncrementSizeListener);
        arConsoleManager.AddConsoleCommand(ARConsoleManager.ConsoleCommandType.DecrementSize, DecrementSizeListener);
        arConsoleManager.AddConsoleCommand(ARConsoleManager.ConsoleCommandType.FlipCameraHorizontal, FlipCameraHorizontalListener);
        arConsoleManager.AddConsoleCommand(ARConsoleManager.ConsoleCommandType.FlipCameraVertical, FlipCameraVerticalListener);
        arConsoleManager.AddConsoleCommand(ARConsoleManager.ConsoleCommandType.RestartGame, RestartGameListener);
        arConsoleManager.AddConsoleCommand(ARConsoleManager.ConsoleCommandType.ToggleSpeaker, ToggleSpeakerListener);

        arConsoleManager.AddConsoleCommand(ARConsoleManager.ConsoleCommandType.ToggleCalibration, PlayButtonSound);
        arConsoleManager.AddConsoleCommand(ARConsoleManager.ConsoleCommandType.IncrementSize, PlayButtonSound);
        arConsoleManager.AddConsoleCommand(ARConsoleManager.ConsoleCommandType.DecrementSize, PlayButtonSound);
        arConsoleManager.AddConsoleCommand(ARConsoleManager.ConsoleCommandType.FlipCameraHorizontal, PlayButtonSound);
        arConsoleManager.AddConsoleCommand(ARConsoleManager.ConsoleCommandType.FlipCameraVertical, PlayButtonSound);
        arConsoleManager.AddConsoleCommand(ARConsoleManager.ConsoleCommandType.RestartGame, PlayButtonSound);
        arConsoleManager.AddConsoleCommand(ARConsoleManager.ConsoleCommandType.ToggleSpeaker, PlayButtonSound);

        // Highlight the toggle calibration button at the beginning to help users at the beginning.
        arConsoleManager.HighlightButton(ARConsoleManager.ConsoleCommandType.ToggleCalibration, true);

        gameManager.EventGameStarted += GameStartedListener;
    }

    private void PlayButtonSound()
    {
        SoundManager.Instance.PlayButtonPressed();
    }

    private void ToggleSpeakerListener()
    {
        // Switches between differenz volume levels. 1 => 0.5 => 0
        if(MusicPlayer.FindInstance().Volume == 1)
        {
            MusicPlayer.FindInstance().Volume = 0.5f;
        }
        else if (MusicPlayer.FindInstance().Volume == 0.5f)
        {
            MusicPlayer.FindInstance().Volume = 0;
        }
        else
        {
            MusicPlayer.FindInstance().Volume = 1;
        }
    }

    private void GameStartedListener()
    {
        // Keep in mind that a new game was requested.
        NewGameRequested = true;
    }

    private void ConsoleActivatedListener()
    {
        gameManager.SwitchToState(this);

        if (!firstActivation)
        {
            SoundManager.Instance.PlayPause();
        }

        firstActivation = false;
    }

    private void ConsoleDeactivatedListener()
    {
        // If a new game was requested and the console is removed, 
        // start preluding a new game through the game manager.
        if (NewGameRequested)
        {
            gameManager.PreludeNewGame();
        }
        // Else: Start preluding a resumed game.
        else
        {
            gameManager.PreludeResumedGame();
        }

        SoundManager.Instance.PlayResume();
    }

    public void ToggleCalibrationListener()
    {
        // Toggle if the origin marker should update the playField
        arManager.UpdateOriginMarker = !arManager.UpdateOriginMarker;
        UpdateGameStartText();
    }

    /**
     * Updates the text indicating which button to press.
     */
    public void UpdateGameStartText()
    {
        if (arManager.UpdateOriginMarker)
        {
            arConsoleManager.HighlightButton(ARConsoleManager.ConsoleCommandType.ToggleCalibration, true);
            gameManager.PlayField.HowToStartText.GetComponent<TextMesh>().text = "Stop calibration to get ready to play.";
        }
        else
        {
            arConsoleManager.HighlightButton(ARConsoleManager.ConsoleCommandType.ToggleCalibration, false);
            gameManager.PlayField.HowToStartText.GetComponent<TextMesh>().text = "Ready? Remove the console to play!";
        }
    }

    private void IncrementSizeListener()
    {
        if(gameManager.PlayField.Scale > 1.5) return;

        gameManager.PlayField.Scale += 0.1f;
    }

    private void DecrementSizeListener()
    {
        if(gameManager.PlayField.Scale < 0.5) return;

        gameManager.PlayField.Scale -= 0.1f;
    }

    private void FlipCameraHorizontalListener()
    {
        flipCamera.ToggleFlipX();
        // Flip texts of playField correspondingly to avoid mirrored texts.
        gameManager.PlayField.FlipTextsX();
    }

    private void FlipCameraVerticalListener()
    {
        flipCamera.ToggleFlipY();
        // Flip texts of playField correspondingly to avoid mirrored texts.
        gameManager.PlayField.FlipTextsY();
    }

    private void RestartGameListener()
    {
        gameManager.InitializeNewGame();
    }

    public override void Activate()
    {
        // Make the right texts visible.
        gameManager.PlayField.GameStartsText.SetActive(false);
        gameManager.PlayField.HowToStartText.SetActive(true);

        // If the game was over, reinitialize a new one so that the user does not have to press the button.
        if (gameManager.GameOver)
        {
            gameManager.InitializeNewGame();
        }
    }

    public override void Deactivate()
    {
    }

    public override void BeforeARUpdate()
    {
    }

    public override void AfterARUpdate()
    {
    }

    private void SwitchToNextState()
    {
    }

    public override bool AcknowledgeNewRecognition(MarkerObject markerObject)
    {
        return MarkerOkay(markerObject);
    }

    public override bool AcknowledgeUpdate(MarkerObject markerObject)
    {
        return MarkerOkay(markerObject);
    }

    public bool MarkerOkay(MarkerObject markerObject)
    {
        return gameManager.DebugAllowNonConsoleMarkersInMenuState || (markerObject.GetComponent<ARConsoleButton>() != null);
    }

    public override void DetectionStarted(MarkerObject markerObject)
    {
        UpdateControlMarkerStatus(markerObject);
    }

    public override void DetectionUpdated(MarkerObject markerObject)
    {
        UpdateControlMarkerStatus(markerObject);
    }

    public override void DetectionEnded(MarkerObject markerObject)
    {
        UpdateControlMarkerStatus(markerObject);
    }

    private void UpdateControlMarkerStatus(MarkerObject markerObject)
    {
        var controlMarker = markerObject.GetComponent<ControlMarker>();
        if (controlMarker == null)
            return;

        controlMarker.UpdateStatus();
    }
}
