/**
 * Copyright (c) 2013 Tobias Mller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using System;
using UnityEngine;
using System.Collections;

/*
 * The FlipCamera script is used to mirror the camera.
 */
public class FlipCamera : MonoBehaviour
{
    [SerializeField] private bool flipX = false;
    [SerializeField] private bool flipY = false;
    [SerializeField] private bool flipZ = false;
    [SerializeField] private bool backfacing = false;

    void Start()
    {
        UpdateBackfacing();
    }

    void OnPreCull()
    {
        if (!enabled)
            return;

        camera.ResetWorldToCameraMatrix();
        camera.ResetProjectionMatrix();
        camera.projectionMatrix = camera.projectionMatrix * Matrix4x4.Scale(new Vector3(flipX ? -1 : 1, flipY ? -1 : 1, flipZ ? -1 : 1));
    }

    void OnPreRender()
    {
        if (!enabled)
            return;

        if (backfacing)
            GL.SetRevertBackfacing(true);
    }

    void OnPostRender()
    {
        if (!enabled)
            return;

        if (backfacing)
            GL.SetRevertBackfacing(false);

        camera.ResetWorldToCameraMatrix();
        camera.ResetProjectionMatrix();
    }

    /**
     * This method has to be called when flipX, flipY or flipZ are changed
     * to correctly show the current camera image to the user.
     */
    public void UpdateBackfacing()
    {
        int count = 0;

        if (flipX)
            count++;

        if (flipY)
            count++;

        if (flipZ)
            count++;

        backfacing = count % 2 == 1;
    }

    public bool FlipX
    {
        get { return flipX; }
        set
        {
            flipX = value;
            UpdateBackfacing();
        }
    }

    public bool FlipY
    {
        get { return flipY; }
        set
        {
            flipY = value;
            UpdateBackfacing();
        }
    }

    public void ToggleFlipX()
    {
        FlipX = !FlipX;
    }

    public void ToggleFlipY()
    {
        FlipY = !FlipY;
    }

    public Vector2 TransformScreenPosition(Vector2 screenPosition)
    {
        if (FlipX)
            screenPosition.x = Screen.width - screenPosition.x;

        if (FlipY)
            screenPosition.y = Screen.height - screenPosition.y;

        return screenPosition;
    }

    public static FlipCamera FindInstance()
    {
        return ARManager.FindInstance().GetComponentInChildren<FlipCamera>();
    }
}
