﻿/**
 * Copyright (c) 2013 Tobias Müller <Tobias_Mueller@gmx.com>, Tobias Wehrum <Tobias.Wehrum@dragonlab.de>
 * 
 * This file is part of Wizard Defense, A Cooperative Augmented Reality Game.
 * 
 * Distributed under the terms of the GNU GPL v3. For the full notice see the file LICENSE.
 */

using UnityEngine;

public class SoundManager : MonoBehaviour
{
    [SerializeField] private AudioClip buttonPressed;
    [SerializeField] private AudioClip pause;
    [SerializeField] private AudioClip resume;

    [SerializeField] private AudioClip[] monsterAppears;
    [SerializeField] private AudioClip monsterDies;
    [SerializeField] private AudioClip monsterEatsWizard;
    [SerializeField] private AudioClip monsterHit;
    [SerializeField] private float monsterHitFullVolumeAtPower = 4;

    [SerializeField] private AudioClip gameStarts;
    [SerializeField] private AudioClip gameOver;
    [SerializeField] private AudioClip gameWon;

    private AudioSource monsterHitAudioSource;

    private void Awake()
    {
        var monsterHitAudioSourceObject = new GameObject("Monster Hit Audio Source");
        monsterHitAudioSourceObject.transform.parent = transform;

        monsterHitAudioSource = monsterHitAudioSourceObject.AddComponent<AudioSource>();
        monsterHitAudioSource.playOnAwake = false;
        monsterHitAudioSource.clip = monsterHit;
        monsterHitAudioSource.loop = true;
        monsterHitAudioSource.volume = 0;
    }

    public float MonsterWasHitByLaserTotalPower
    {
        get { return monsterHitAudioSource.volume; }

        set
        {
            var volume = Mathf.Clamp01(value / monsterHitFullVolumeAtPower);

            if (monsterHitAudioSource.volume == volume)
                return;

            if (volume > 0)
            {
                if (!monsterHitAudioSource.isPlaying)
                    monsterHitAudioSource.Play();

                monsterHitAudioSource.volume = volume;
            }
            else
            {
                if (monsterHitAudioSource.isPlaying)
                    monsterHitAudioSource.Pause();

                monsterHitAudioSource.volume = 0;
            }
        }
    }

    public void PlayButtonPressed()
    {
        PlaySound(buttonPressed);
    }

    public void PlayPause()
    {
        PlaySound(pause);
    }

    public void PlayResume()
    {
        PlaySound(resume);
    }

    public void PlayMonsterAppears(int healthIndex)
    {
        PlaySound(monsterAppears[healthIndex]);
    }

    public void PlayMonsterDies()
    {
        PlaySound(monsterDies);
    }

    public void PlayMonsterEatsWizard()
    {
        PlaySound(monsterEatsWizard);
    }

    public void PlayGameStarts()
    {
        PlaySound(gameStarts);
    }

    public void PlayGameOver()
    {
        PlaySound(gameOver);
    }

    public void PlayGameWon()
    {
        PlaySound(gameWon);
    }

    private static void PlaySound(AudioClip[] audioClips, Vector3? position = null)
    {
        if ((audioClips == null) || (audioClips.Length == 0))
            return;

        var audioClip = audioClips[UnityEngine.Random.Range(0, audioClips.Length)];

        PlaySound(audioClip, position);
    }

    private static void PlaySound(AudioClip audioClip, Vector3? position = null)
    {
        if (audioClip == null)
            return;

        if (position == null)
            position = Camera.main.transform.position;

        audioClip.PlayAtPosition(position.Value);
    }

    public static SoundManager Instance { get { return UnityHelper.FindFirstActiveInstance<SoundManager>(); } }
}
